"""The DatasetDescriptor class."""

from typing import List

from .model_attribute import model_attribute
from glotaran.parameter import Parameter


@model_attribute(properties={
    'megacomplex': List[str],
    'scale': {'type': Parameter, 'default': None, 'allow_none': True},
})
class DatasetDescriptor:
    """A `DatasetDescriptor` describes a dataset in terms of a glotaran model.
    It contains references to model items which describe the physical model for
    a given dataset.

    A general dataset describtor assigns one or more megacomplexes and a scale
    parameter.
    """



import numpy as np

from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_image import KineticImageModel
from glotaran.builtin.models.kinetic_image.kinetic_image_matrix import kinetic_image_matrix


def test_baseline():
    model = KineticImageModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['2']
            },

        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {("s1", "s1"): '1', }}
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
                'baseline': True,
            },
        },
    })

    parameter = ParameterGroup.from_list([
        101e-4,
        [1, {'vary': False, 'non-negative': False}],
        [42, {'vary': False, 'non-negative': False}],
    ])

    time = np.asarray(np.arange(0, 50, 1.5))
    dataset = model.dataset['dataset1'].fill(model, parameter)
    compartments, matrix = kinetic_image_matrix(dataset, time, 0)

    assert len(compartments) == 2
    assert compartments[1] == 'dataset1_baseline'

    assert matrix.shape == (time.size, 2)
    assert np.all(matrix[:, 1] == 1)



from glotaran.model import Model, model, model_attribute

from glotaran.parse import register


def mock_matrix_fun():
    pass


@model_attribute()
class MockMegacomplex:
    pass


@model('register_mock',
       matrix=mock_matrix_fun,
       megacomplex_type=MockMegacomplex,
       )
class MockModel(Model):
    pass


def test_register():
    assert register.known_model('register_mock')
    assert issubclass(register.get_model('register_mock'), MockModel)



import collections
import dask
import lmfit
import numpy as np

from glotaran.parameter import ParameterGroup

from . import problem_bag, residual_calculation
from .matrix_calculation import (
    calculate_index_independend_ungrouped_matrices,
    calculate_index_independend_grouped_matrices,
    create_index_dependend_ungrouped_matrix_jobs,
    create_index_dependend_grouped_matrix_jobs
)
from .nnls import residual_nnls
from .variable_projection import residual_variable_projection
from .result import Result

ResultFuture = \
    collections.namedtuple('ResultFuture', 'bag clp_label matrix full_clp_label clp residual')


def optimize(scheme, verbose=True, client=None):

    initial_parameter = scheme.parameter.as_parameter_dict()

    if client is not None:
        scheme = client.scatter(scheme)
        optimization_result_future = \
            client.submit(optimize_task, initial_parameter, scheme, verbose)
        result = optimization_result_future.result()

    else:
        result = optimize_task(initial_parameter, scheme, verbose)

    return result


def optimize_task(initial_parameter, scheme, verbose):

    problem_bag, groups = _create_problem_bag(scheme)

    minimizer = lmfit.Minimizer(
        calculate_penalty,
        initial_parameter,
        fcn_args=[scheme, problem_bag, groups],
        fcn_kws=None,
        iter_cb=None,
        scale_covar=True,
        nan_policy='omit',
        reduce_fcn=None,
        **{})
    verbose = 2 if verbose else 0
    lm_result = minimizer.minimize(
        method='least_squares', verbose=verbose, max_nfev=scheme.nfev)

    parameter = ParameterGroup.from_parameter_dict(lm_result.params)
    datasets = _create_result(scheme, parameter)
    covar = lm_result.covar if hasattr(lm_result, 'covar') else None

    return Result(scheme, datasets, parameter,
                  lm_result.nfev, lm_result.nvarys, lm_result.ndata, lm_result.nfree,
                  lm_result.chisqr, lm_result.redchi, lm_result.var_names, covar)


def calculate_penalty(parameter, scheme, bag, groups):
    parameter = ParameterGroup.from_parameter_dict(parameter)
    residual_function = residual_nnls if scheme.nnls else residual_variable_projection
    if scheme.model.grouped():
        if scheme.model.index_dependend():
            _, _, constraint_labels_and_matrices = \
                create_index_dependend_grouped_matrix_jobs(
                    scheme, bag, parameter
                )
            _, _, _, penalty = \
                residual_calculation.create_index_dependend_grouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
        else:

            _, _, constraint_labels_and_matrices = \
                calculate_index_independend_grouped_matrices(scheme, groups, parameter)

            _, _, _, penalty = \
                residual_calculation.create_index_independend_grouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
    else:
        if scheme.model.index_dependend():
            _, _, constraint_labels_and_matrices = \
                create_index_dependend_ungrouped_matrix_jobs(
                    scheme, bag, parameter
                )
            _, _, _, penalty = \
                residual_calculation.create_index_dependend_ungrouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
        else:

            _, _, constraint_labels_and_matrices = \
                calculate_index_independend_ungrouped_matrices(scheme, parameter)

            _, _, _, penalty = \
                residual_calculation.create_index_independend_ungrouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
    penalty = penalty.compute()
    return penalty


def _create_problem_bag(scheme):
    groups = None
    if scheme.model.grouped():
        bag, groups = problem_bag.create_grouped_bag(scheme)
        bag = bag.persist()
    else:
        bag = problem_bag.create_ungrouped_bag(scheme)
    return bag, groups


def _create_result(scheme, parameter):

    residual_function = residual_nnls if scheme.nnls else residual_variable_projection
    model = scheme.model
    datasets = scheme.data

    if model.grouped():
        bag, groups = problem_bag.create_grouped_bag(scheme)

        if model.index_dependend():
            clp_labels, matrices, constraint_labels_and_matrices = \
                create_index_dependend_grouped_matrix_jobs(
                    scheme, bag, parameter
                )
            reduced_clp_labels, reduced_clps, residuals, _ = \
                residual_calculation.create_index_dependend_grouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
        else:
            clp_labels, matrices, constraint_labels_and_matrices = \
                calculate_index_independend_grouped_matrices(
                    scheme, groups, parameter
                )
            reduced_clp_labels, reduced_clps, residuals, _ = \
                residual_calculation.create_index_independend_grouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
    else:
        bag = problem_bag.create_ungrouped_bag(scheme)

        if model.index_dependend():
            clp_labels, matrices, constraint_labels_and_matrices = \
                create_index_dependend_ungrouped_matrix_jobs(
                    scheme, bag, parameter
                )
            reduced_clp_labels, reduced_clps, residuals, _ = \
                residual_calculation.create_index_dependend_ungrouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )
        else:
            clp_labels, matrices, constraint_labels_and_matrices = \
                calculate_index_independend_ungrouped_matrices(
                    scheme, parameter
                )
            reduced_clp_labels, reduced_clps, residuals, _ = \
                residual_calculation.create_index_independend_ungrouped_residual(
                    scheme, parameter, bag, constraint_labels_and_matrices, residual_function
                )

    indices = None

    if model.grouped():
        indices = bag.map(lambda group: [d.index for d in group.descriptor])
        if model.index_dependend():
            groups, indices, clp_labels, matrices, reduced_clp_labels, reduced_clps, residuals = \
                    dask.compute(groups, indices, clp_labels, matrices,
                                 reduced_clp_labels, reduced_clps, residuals)
        else:
            groups, indices, reduced_clp_labels, reduced_clps, residuals = dask.compute(
                groups, indices, reduced_clp_labels, reduced_clps, residuals)

    for label, dataset in datasets.items():
        if model.grouped():
            if model.index_dependend():
                groups = bag.map(lambda group: [d.dataset for d in group.descriptor]).compute()
                for i, group in enumerate(groups):
                    if label in group:
                        group_index = group.index(label)
                        if 'matrix' not in dataset:
                            # we assume that the labels are the same, this might not be true in
                            # future models
                            dataset.coords['clp_label'] = clp_labels[i][group_index]

                            dim1 = dataset.coords[model.global_dimension].size
                            dim2 = dataset.coords[model.matrix_dimension].size
                            dim3 = dataset.clp_label.size
                            dataset['matrix'] = \
                                (((model.global_dimension),
                                  (model.matrix_dimension),
                                  ('clp_label')),
                                 np.zeros((dim1, dim2, dim3), dtype=np.float64))
                        dataset.matrix.loc[{model.global_dimension: indices[i][group_index]}] = \
                            matrices[i][group_index]
            else:
                clp_label, matrix = dask.compute(
                    clp_labels[label],
                    matrices[label],
                )
                dataset.coords['clp_label'] = clp_label
                dataset['matrix'] = (((model.matrix_dimension), ('clp_label')), matrix)
            dim1 = dataset.coords[model.global_dimension].size
            dim2 = dataset.coords['clp_label'].size
            dataset['clp'] = ((model.global_dimension, 'clp_label'),
                              np.zeros((dim1, dim2), dtype=np.float64))

            dim1 = dataset.coords[model.matrix_dimension].size
            dim2 = dataset.coords[model.global_dimension].size
            dataset['residual'] = ((model.matrix_dimension, model.global_dimension),
                                   np.zeros((dim1, dim2), dtype=np.float64))
            idx = 0
            for i, group in enumerate(groups):
                if label in group:
                    index = indices[i][group.index(label)]
                    for j, clp in enumerate(reduced_clp_labels[i]):
                        if clp in dataset.clp_label:
                            dataset.clp.loc[{'clp_label': clp, model.global_dimension: index}] = \
                                    reduced_clps[i][j]
                    start = 0
                    for dset in group:
                        if dset == label:
                            break
                        start += datasets[dset].coords[model.matrix_dimension].size
                    end = start + dataset.coords[model.matrix_dimension].size
                    dataset.residual.loc[{model.global_dimension: index}] = residuals[i][start:end]

        else:
            clp_label, matrix, reduced_clp_label, reduced_clp, residual = dask.compute(
                clp_labels[label],
                matrices[label],
                reduced_clp_labels[label],
                reduced_clps[label],
                residuals[label],
            )
            reduced_clp = np.asarray(reduced_clp)

            if model.index_dependend():
                # we assume that the labels are the same, this might not be true in future models
                dataset.coords['clp_label'] = clp_label[0]
                dataset['matrix'] = \
                    (((model.global_dimension), (model.matrix_dimension), ('clp_label')), matrix)
            else:
                dataset.coords['clp_label'] = clp_label
                dataset['matrix'] = (((model.matrix_dimension), ('clp_label')), matrix)

            dim1 = dataset.coords[model.global_dimension].size
            dim2 = dataset.coords['clp_label'].size
            dataset['clp'] = ((model.global_dimension, 'clp_label'),
                              np.zeros((dim1, dim2), dtype=np.float64))
            for i, clp in enumerate(reduced_clp_label):
                if model.index_dependend():
                    idx = dataset.coords[model.global_dimension][i]
                    for c in clp:
                        if c not in reduced_clp_label[i]:
                            continue
                        j = reduced_clp_label[i].index(c)
                        dataset.clp.loc[{'clp_label': c, model.global_dimension: idx}] = \
                            reduced_clp[i][j]
                else:
                    dataset.clp.loc[{'clp_label': clp}] = \
                            reduced_clp[:, i]

            dataset['residual'] = (((model.matrix_dimension), (model.global_dimension)),
                                   np.asarray(residual).T)

    if 'weight' in dataset:
        dataset['weighted_residual'] = dataset.residual
        dataset.residual = np.multiply(dataset.weighted_residual, dataset.weight**-1)

    size = dataset.residual.shape[0] * dataset.residual.shape[1]
    dataset.attrs['root_mean_square_error'] = \
        np.sqrt((dataset.residual**2).sum()/size).values

    l, v, r = np.linalg.svd(dataset.residual)

    dataset['residual_left_singular_vectors'] = \
        ((model.matrix_dimension, 'left_singular_value_index'), l)

    dataset['residual_right_singular_vectors'] = \
        (('right_singular_value_index', model.global_dimension), r)

    dataset['residual_singular_values'] = \
        (('singular_value_index'), v)

    # reconstruct fitted data

    dataset['fitted_data'] = dataset.data - dataset.residual

    if callable(model.finalize_data):
        model.finalize_data(indices, reduced_clp_labels, reduced_clps, parameter, datasets)

    return datasets



import pytest

from glotaran.parameter import ParameterGroup
from glotaran.parse.parser import load_yml_file

from glotaran.builtin.models.kinetic_image.initial_concentration \
        import InitialConcentration
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.builtin.models.kinetic_image.kinetic_image_megacomplex \
        import KineticImageMegacomplex

from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_dataset_descriptor \
    import KineticSpectrumDatasetDescriptor
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model \
        import KineticSpectrumModel
from glotaran.builtin.models.kinetic_spectrum.spectral_constraints \
    import ZeroConstraint
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties \
    import EqualAreaPenalty
from glotaran.builtin.models.kinetic_spectrum.spectral_shape \
        import SpectralShapeGaussian

from os.path import join, dirname, abspath
import numpy as np

THIS_DIR = dirname(abspath(__file__))


@pytest.fixture
def model():
    spec_path = join(THIS_DIR, 'test_model_spec_kinetic.yml')
    m = load_yml_file(spec_path)
    print(m.markdown())
    return m


def test_correct_model(model):
    assert type(model).__name__ == 'KineticSpectrumModel'
    assert isinstance(model, KineticSpectrumModel)


def test_dataset(model):
    assert len(model.dataset) == 2

    assert 'dataset1' in model.dataset
    dataset = model.dataset['dataset1']
    assert isinstance(dataset, KineticSpectrumDatasetDescriptor)
    assert dataset.label == 'dataset1'
    assert dataset.megacomplex == ["cmplx1"]
    assert dataset.initial_concentration == "inputD1"
    assert dataset.irf == "irf1"
    assert dataset.scale == 1

    assert len(dataset.shape) == 2
    assert dataset.shape['s1'] == "shape1"
    assert dataset.shape['s2'] == "shape2"

    dataset = model.dataset['dataset2']


def test_spectral_constraints(model):
    print(model.spectral_constraints)
    assert len(model.spectral_constraints) == 2

    assert any(isinstance(c, ZeroConstraint) for c in
               model.spectral_constraints)

    zcs = [zc for zc in model.spectral_constraints
           if zc.type == 'zero']
    assert len(zcs) == 2
    for zc in zcs:
        assert zc.compartment == 's1'
        assert zc.interval == [(1, 100), (2, 200)]


def test_spectral_penalties(model):
    assert len(model.equal_area_penalties) == 1
    assert all(isinstance(c, EqualAreaPenalty) for c in
               model.equal_area_penalties)
    eac = model.equal_area_penalties[0]
    assert eac.compartment == 's3'
    assert eac.interval == [(670, 810)]
    assert eac.target == 's2'
    assert eac.parameter == 55
    assert eac.weight == 0.0016


def test_spectral_relations(model):
    print(model.spectral_relations)
    assert len(model.spectral_relations) == 1

    rel = model.spectral_relations[0]

    assert rel.compartment == 's1'
    assert rel.target == 's2'
    assert rel.interval == [(1, 100), (2, 200)]


def test_initial_concentration(model):
    assert len(model.initial_concentration) == 2

    i = 1
    for _ in model.initial_concentration:
        label = "inputD{}".format(i)
        assert label in model.initial_concentration
        initial_concentration = model.initial_concentration[label]
        assert initial_concentration.compartments == ['s1', 's2', 's3']
        assert isinstance(initial_concentration, InitialConcentration)
        assert initial_concentration.label == label
        assert initial_concentration.parameters == [1, 2, 3]


def test_irf(model):
    assert len(model.irf) == 2

    i = 1
    for _ in model.irf:
        label = "irf{}".format(i)
        assert label in model.irf
        irf = model.irf[label]
        assert isinstance(irf, IrfMultiGaussian)
        assert irf.label == label
        want = [1] if i == 1 else [1, 2]
        assert irf.center == want
        want = [2] if i == 1 else [3, 4]
        assert irf.width == want
        want = [3] if i == 1 else [5, 6]
        if i == 2:
            assert irf.center_dispersion == want
            want = [7, 8]
            assert irf.width_dispersion == want
            want = [9]
            assert irf.scale == want
        assert not irf.normalize

        if i == 2:
            assert irf.backsweep
            assert irf.backsweep_period, 55
        else:
            assert not irf.backsweep
            assert irf.backsweep_period is None

        i = i + 1


def test_k_matrices(model):
    assert "km1" in model.k_matrix
    parameter = ParameterGroup.from_list([1, 2, 3, 4, 5, 6, 7])
    reduced = model.k_matrix["km1"].fill(model, parameter).reduced(['s1', 's2', 's3', 's4'])
    assert np.array_equal(reduced,
                          np.asarray([[1, 3, 5, 7],
                                      [2, 0, 0, 0],
                                      [4, 0, 0, 0],
                                      [6, 0, 0, 0]]))


def test_shapes(model):

    assert "shape1" in model.shape

    shape = model.shape["shape1"]
    assert isinstance(shape, SpectralShapeGaussian)
    assert shape.amplitude.full_label == "shape.1"
    assert shape.location.full_label == "shape.2"
    assert shape.width.full_label == "shape.3"


def test_megacomplexes(model):
    assert len(model.megacomplex) == 3

    i = 1
    for _ in model.megacomplex:
        label = "cmplx{}".format(i)
        assert label in model.megacomplex
        megacomplex = model.megacomplex[label]
        assert isinstance(megacomplex, KineticImageMegacomplex)
        assert megacomplex.label == label
        assert megacomplex.k_matrix == ["km{}".format(i)]
        i = i + 1



import typing
import numpy as np

from glotaran.model import model
from glotaran.parameter import ParameterGroup

from glotaran.builtin.models.kinetic_image.kinetic_image_megacomplex import KineticImageMegacomplex
from glotaran.builtin.models.kinetic_image.kinetic_image_model import KineticImageModel

from .kinetic_spectrum_dataset_descriptor import KineticSpectrumDatasetDescriptor
from .kinetic_spectrum_result import finalize_kinetic_spectrum_result
from .kinetic_spectrum_matrix import kinetic_spectrum_matrix
from .spectral_constraints import SpectralConstraint, apply_spectral_constraints
from .spectral_irf import IrfSpectralMultiGaussian
from .spectral_matrix import spectral_matrix
from .spectral_penalties import EqualAreaPenalty, has_spectral_penalties, apply_spectral_penalties
from .spectral_relations import SpectralRelation, apply_spectral_relations
from .spectral_shape import SpectralShape


def has_kinetic_model_constraints(model: typing.Type['KineticModel']) -> bool:
    return len(model.spectral_relations) + len(model.spectral_constraints) != 0


def apply_kinetic_model_constraints(
        model: typing.Type['KineticModel'],
        parameter: ParameterGroup,
        clp_labels: typing.List[str],
        matrix: np.ndarray,
        index: float):
    clp_labels, matrix = apply_spectral_relations(model, parameter, clp_labels, matrix, index)
    clp_labels, matrix = apply_spectral_constraints(model, clp_labels, matrix, index)
    return clp_labels, matrix


def index_dependend(model: typing.Type['KineticModel']):
    if any([
        isinstance(irf, IrfSpectralMultiGaussian) and irf.dispersion_center is not None
        for irf in model.irf.values()
    ]):
        return True
    if len(model.spectral_relations) != 0:
        return True
    if len(model.spectral_constraints) != 0:
        return True
    return False


def grouped(model: typing.Type['KineticModel']):
    return len(model.dataset) != 1


@model(
    'kinetic-spectrum',
    attributes={
        'equal_area_penalties': EqualAreaPenalty,
        'shape': SpectralShape,
        'spectral_constraints': SpectralConstraint,
        'spectral_relations': SpectralRelation,
    },
    dataset_type=KineticSpectrumDatasetDescriptor,
    megacomplex_type=KineticImageMegacomplex,
    matrix=kinetic_spectrum_matrix,
    matrix_dimension='time',
    global_matrix=spectral_matrix,
    global_dimension='spectral',
    has_matrix_constraints_function=has_kinetic_model_constraints,
    constrain_matrix_function=apply_kinetic_model_constraints,
    has_additional_penalty_function=has_spectral_penalties,
    additional_penalty_function=apply_spectral_penalties,
    grouped=grouped,
    index_dependend=index_dependend,
    finalize_data_function=finalize_kinetic_spectrum_result,
)
class KineticSpectrumModel(KineticImageModel):
    pass



import os

DATA_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "data_files"))

LEGACY_FILES = {"flim_file": os.path.join(DATA_DIR, "FLIM.sdt.ascii"),
                "flim_traces": os.path.join(DATA_DIR, "FLIM_legacy_trace.ssv"),
                "flim_map": os.path.join(DATA_DIR, "FLIM_legacy_sum_map.ssv")}

TEMPORAL_DATA = {"sdt": os.path.join(DATA_DIR, "temporal.sdt"),
                 "csv": os.path.join(DATA_DIR, "temporal.sdt.kin")}

FLIM_DATA = {"sdt": os.path.join(DATA_DIR, "FLIM.sdt"),
             "csv": os.path.join(DATA_DIR, "FLIM.sdt.ascii")}



import pytest
import numpy as np

from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel


class OneComponentOneChannel:
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['2']
            },

        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {("s1", "s1"): '1', }}
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
            },
        },
    })
    sim_model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['2']
            },
        },
        'shape': {'sh1': ['one']},
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {("s1", "s1"): '1', }}
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
                'shape': {'s1': 'sh1'}
            },
        },
    })

    initial = ParameterGroup.from_list([101e-4, [1, {'vary': False, 'non-negative': False}]])
    wanted = ParameterGroup.from_list([101e-3, [1, {'vary': False, 'non-negative': False}]])

    time = np.asarray(np.arange(0, 50, 1.5))
    spectral = np.asarray([0])
    axis = {"time": time, "spectral": spectral}


class OneComponentOneChannelGaussianIrf:
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['2']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {("s1", "s1"): '1', }}
        },
        'irf': {
            'irf1': {'type': 'spectral-gaussian', 'center': '2', 'width': '3'},
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
            },
        },
    })
    sim_model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['4']
            },
        },
        'shape': {'sh1': ['one']},
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {("s1", "s1"): '1', }}
        },
        'irf': {
            'irf1': {'type': 'spectral-gaussian', 'center': '2', 'width': '3'},
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
                'shape': {'s1': 'sh1'}
            },
        },
    })

    initial = ParameterGroup.from_list([101e-4, 0.1, 5,
                                        [1, {'vary': False, 'non-negative': False}]])
    wanted = ParameterGroup.from_list([101e-3, 0.3, 10,
                                       [1, {'vary': False, 'non-negative': False}]])

    time = np.asarray(np.arange(-10, 50, 1.5))
    spectral = np.asarray([0])
    axis = {"time": time, "spectral": spectral}


class ThreeComponentParallel:
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['j.1', 'j.1', 'j.1']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s2", "s1"): 'kinetic.1',
                ("s3", "s2"): 'kinetic.2',
                ("s3", "s3"): 'kinetic.3',
            }}
        },
        'irf': {
            'irf1': {
                'type': 'spectral-multi-gaussian',
                'center': ['irf.center'],
                'width': ['irf.width']},
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
            },
        },
    })
    sim_model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['j.1', 'j.1', 'j.1']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
                ("s2", "s2"): 'kinetic.2',
                ("s3", "s3"): 'kinetic.3',
            }}
        },
        'shape': {
            'sh1': {
                'type': "gaussian",
                'amplitude': "shape.amps.1",
                'location': "shape.locs.1",
                'width': "shape.width.1",
            },
            'sh2': {
                'type': "gaussian",
                'amplitude': "shape.amps.2",
                'location': "shape.locs.2",
                'width': "shape.width.2",
            },
            'sh3': {
                'type': "gaussian",
                'amplitude': "shape.amps.3",
                'location': "shape.locs.3",
                'width': "shape.width.3",
            },
        },
        'irf': {
            'irf1': {
                'type': 'spectral-multi-gaussian',
                'center': ['irf.center'],
                'width': ['irf.width']
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
                'shape': {'s1': 'sh1', 's2': 'sh2', 's3': 'sh3'}
            },
        },
    })

    initial = ParameterGroup.from_dict({
        'kinetic': [
            ["1", 300e-3],
            ["2", 500e-4],
            ["3", 700e-5],
        ],
        'irf': [['center', 1.3], ['width', 7.8]],
        'j': [['1', 1, {'vary': False, 'non-negative': False}]],
    })
    wanted = ParameterGroup.from_dict({
        'kinetic': [
            ["1", 301e-3],
            ["2", 502e-4],
            ["3", 705e-5],
        ],
        'shape': {'amps': [7, 3, 30], 'locs': [620, 670, 720], 'width': [10, 30, 50]},
        'irf': [['center', 1.3], ['width', 7.8]],
        'j': [['1', 1, {'vary': False, 'non-negative': False}]],
    })

    time = np.arange(-10, 100, 1.5)
    spectral = np.arange(600, 750, 10)
    axis = {"time": time, "spectral": spectral}


class ThreeComponentSequential:
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['j.1', 'j.0', 'j.0']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s2", "s1"): 'kinetic.1',
                ("s3", "s2"): 'kinetic.2',
                ("s3", "s3"): 'kinetic.3',
            }}
        },
        'irf': {
            'irf1': {
                'type': 'spectral-multi-gaussian',
                'center': ['irf.center'],
                'width': ['irf.width']
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
            },
        },
    })
    sim_model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['j.1', 'j.0', 'j.0']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s2", "s1"): 'kinetic.1',
                ("s3", "s2"): 'kinetic.2',
                ("s3", "s3"): 'kinetic.3',
            }}
        },
        'shape': {
            'sh1': {
                'type': "gaussian",
                'amplitude': "shape.amps.1",
                'location': "shape.locs.1",
                'width': "shape.width.1",
            },
            'sh2': {
                'type': "gaussian",
                'amplitude': "shape.amps.2",
                'location': "shape.locs.2",
                'width': "shape.width.2",
            },
            'sh3': {
                'type': "gaussian",
                'amplitude': "shape.amps.3",
                'location': "shape.locs.3",
                'width': "shape.width.3",
            },
        },
        'irf': {
            'irf1': {'type': 'spectral-multi-gaussian',
                     'center': ['irf.center'],
                     'width': ['irf.width']},
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
                'shape': {'s1': 'sh1', 's2': 'sh2', 's3': 'sh3'}
            },
        },
    })

    initial = ParameterGroup.from_dict({
        'kinetic': [
            ["1", 501e-3],
            ["2", 202e-4],
            ["3", 105e-5],
            {'non-negative': True},
        ],
        'irf': [['center', 1.3], ['width', 7.8]],
        'j': [['1', 1, {'vary': False, 'non-negative': False}],
              ['0', 0, {'vary': False, 'non-negative': False}]],
    })
    wanted = ParameterGroup.from_dict({
        'kinetic': [
            ["1", 501e-3],
            ["2", 202e-4],
            ["3", 105e-5],
        ],
        'shape': {'amps': [3, 1, 5], 'locs': [620, 670, 720], 'width': [10, 30, 50]},
        'irf': [['center', 1.3], ['width', 7.8]],
        'j': [['1', 1, {'vary': False, 'non-negative': False}],
              ['0', 0, {'vary': False, 'non-negative': False}]],
    })

    time = np.asarray(np.arange(-10, 50, 1.0))
    spectral = np.arange(600, 750, 5.0)
    axis = {"time": time, "spectral": spectral}


@pytest.mark.parametrize("suite", [
    OneComponentOneChannel,
    OneComponentOneChannelGaussianIrf,
    ThreeComponentParallel,
    ThreeComponentSequential,
])
@pytest.mark.parametrize("nnls", [True, False])
def test_kinetic_model(suite, nnls):

    model = suite.model
    print(model.validate())
    assert model.valid()

    sim_model = suite.sim_model
    print(sim_model.validate())
    assert sim_model.valid()

    wanted = suite.wanted
    print(sim_model.validate(wanted))
    print(wanted)
    assert sim_model.valid(wanted)

    initial = suite.initial
    print(model.validate(initial))
    assert model.valid(initial)

    print(model.markdown(wanted))

    dataset = sim_model.simulate('dataset1', wanted, suite.axis)

    assert dataset.data.shape == \
        (suite.axis['time'].size, suite.axis['spectral'].size)

    data = {'dataset1': dataset}

    result = model.optimize(initial, data, nnls=nnls, max_nfev=20)
    print(result.optimized_parameter)

    for label, param in result.optimized_parameter.all():
        assert np.allclose(param.value, wanted.get(label).value,
                           rtol=1e-1)

    resultdata = result.data["dataset1"]

    print(resultdata)

    assert np.array_equal(dataset['time'], resultdata['time'])
    assert np.array_equal(dataset['spectral'], resultdata['spectral'])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)

    assert 'species_associated_spectra' in resultdata
    assert 'decay_associated_spectra' in resultdata



import numpy as np
import xarray as xr
from copy import deepcopy

from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel


def test_spectral_penalties():

    model_without_penalty = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['i.1', 'i.2', 'i.3']
            },

        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
                ("s2", "s2"): 'kinetic.1',
                ("s3", "s3"): 'kinetic.1',
            }}
        },
        'spectral_relations': [
            {
                'compartment': 's1',
                'target': 's2',
                'parameter': 'rel.1',
                'interval': [(0, 2)],
            },
        ],
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
            },
        },
    })

    weight = 0.1
    model_with_penalty = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2', 's3'],
                'parameters': ['i.1', 'i.2', 'i.3']
            },

        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
                ("s2", "s2"): 'kinetic.1',
                ("s3", "s3"): 'kinetic.1',
            }}
        },
        'equal_area_penalties': [
            {
                'compartment': 's2',
                'target': 's3',
                'parameter': 'pen.1',
                'interval': [(0, 2)],
                'weight': weight
            },
        ],
        'spectral_relations': [
            {
                'compartment': 's1',
                'target': 's2',
                'parameter': 'rel.1',
                'interval': [(0, 2)],
            },
        ],
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
            },
        },
    })
    print(model_with_penalty)

    rel1 = 2
    pen = 0.5
    parameter = ParameterGroup.from_dict({
        'kinetic': [1e-4],
        'i': [1, 1, 1],
        'rel': [rel1],
        'pen': [pen],
    })

    time = np.asarray(np.arange(0, 50, 1.5))
    clp = xr.DataArray([[1., 2., 4]],
                       coords=(('spectral', [1]), ('clp_label', ['s1', 's2', 's3'])))

    data = model_without_penalty.simulate('dataset1', parameter, clp=clp,
                                          axes={'time': time, 'spectral': np.array([1])})

    result_with_penalty = \
        model_with_penalty.optimize(parameter, {'dataset1': data}, max_nfev=1)

    parameter_no_penalty = deepcopy(parameter)
    del(parameter_no_penalty['pen'])

    result_without_penalty = \
        model_without_penalty.optimize(parameter_no_penalty, {'dataset1': data}, max_nfev=1)

    result_data = result_with_penalty.data['dataset1']
    wanted_penalty = result_data.species_associated_spectra.sel(species='s2') - \
        result_data.species_associated_spectra.sel(species='s3') * pen
    wanted_penalty *= weight
    wanted_penalty **= 2
    wanted_penalty = np.sum(wanted_penalty.values)

    additional_penalty = result_with_penalty.chisqr - result_without_penalty.chisqr
    assert np.isclose(additional_penalty, wanted_penalty)


if __name__ == "__main__":
    test_spectral_penalties()



import pytest
import numpy as np
import xarray as xr

from glotaran import ParameterGroup
from glotaran.builtin.models.doas import DOASModel
from glotaran.builtin.models.doas.doas_matrix import calculate_doas_matrix


class OneOscillation():
    sim_model = DOASModel.from_dict({
        'oscillation': {
            'osc1': {'frequency': 'osc.freq', 'rate': 'osc.rate'}
        },
        'megacomplex': {
            'm1': {'oscillation': ['osc1']}
        },
        'shape': {
            'sh1': {
                'type': "gaussian",
                'amplitude': "shape.amps.1",
                'location': "shape.locs.1",
                'width': "shape.width.1",
            },
        },
        'dataset': {
            'dataset1': {
                'megacomplex': ['m1'],
                'shape': {'osc1': 'sh1'}
            }
        }
    })

    model = DOASModel.from_dict({
        'oscillation': {
            'osc1': {'frequency': 'osc.freq', 'rate': 'osc.rate'}
        },
        'megacomplex': {
            'm1': {'oscillation': ['osc1']}
        },
        'dataset': {
            'dataset1': {
                'megacomplex': ['m1']
            }
        }
    })

    wanted_parameter = ParameterGroup.from_dict({
        'osc': [
            ['freq', 25.5],
            ['rate', 0.1],
        ],
        'shape': {'amps': [7], 'locs': [5], 'width': [4]},
    })

    parameter = ParameterGroup.from_dict({
        'osc': [
            ['freq', 20],
            ['rate', 0.3],
        ],
    })

    time = np.arange(0, 3, 0.01)
    spectral = np.arange(0, 10)
    axis = {'time': time, 'spectral': spectral}

    wanted_clp = ['osc1_cos', 'osc1_sin']
    wanted_shape = (300, 2)


class OneOscillationWithIrf():
    sim_model = DOASModel.from_dict({
        'oscillation': {
            'osc1': {'frequency': 'osc.freq', 'rate': 'osc.rate'}
        },
        'megacomplex': {
            'm1': {'oscillation': ['osc1']}
        },
        'shape': {
            'sh1': {
                'type': "gaussian",
                'amplitude': "shape.amps.1",
                'location': "shape.locs.1",
                'width': "shape.width.1",
            },
        },
        'irf': {
            'irf1': {'type': 'gaussian', 'center': 'irf.center', 'width': 'irf.width'},
        },
        'dataset': {
            'dataset1': {
                'megacomplex': ['m1'],
                'shape': {'osc1': 'sh1'},
                'irf': 'irf1',
            }
        }
    })

    model = DOASModel.from_dict({
        'oscillation': {
            'osc1': {'frequency': 'osc.freq', 'rate': 'osc.rate'}
        },
        'megacomplex': {
            'm1': {'oscillation': ['osc1']}
        },
        'irf': {
            'irf1': {'type': 'gaussian', 'center': 'irf.center', 'width': 'irf.width'},
        },
        'dataset': {
            'dataset1': {
                'megacomplex': ['m1'],
                'irf': 'irf1',
            }
        }
    })

    wanted_parameter = ParameterGroup.from_dict({
        'osc': [
            ['freq', 25],
            ['rate', 0.1],
        ],
        'shape': {'amps': [7], 'locs': [5], 'width': [4]},
        'irf': [['center', 0.3], ['width', 0.1]],
    })

    parameter = ParameterGroup.from_dict({
        'osc': [
            ['freq', 25],
            ['rate', 0.1],
        ],
        'irf': [['center', 0.3], ['width', 0.1]],
    })

    time = np.arange(0, 3, 0.01)
    spectral = np.arange(0, 10)
    axis = {'time': time, 'spectral': spectral}

    wanted_clp = ['osc1_cos', 'osc1_sin']
    wanted_shape = (300, 2)


class OneOscillationWithSequentialModel():
    sim_model = DOASModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2'],
                'parameters': ['j.1', 'j.0']
            },
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s2", "s1"): 'kinetic.1',
                ("s2", "s2"): 'kinetic.2',
            }}
        },
        'oscillation': {
            'osc1': {'frequency': 'osc.freq', 'rate': 'osc.rate'}
        },
        'megacomplex': {
            'm1': {
                'k_matrix': ['k1'],
                'oscillation': ['osc1'],
            }
        },
        'shape': {
            'sh1': {
                'type': "gaussian",
                'amplitude': "shape.amps.1",
                'location': "shape.locs.1",
                'width': "shape.width.1",
            },
            'sh2': {
                'type': "gaussian",
                'amplitude': "shape.amps.2",
                'location': "shape.locs.2",
                'width': "shape.width.2",
            },
            'sh3': {
                'type': "gaussian",
                'amplitude': "shape.amps.3",
                'location': "shape.locs.3",
                'width': "shape.width.3",
            },
        },
        'irf': {
            'irf1': {'type': 'gaussian', 'center': 'irf.center', 'width': 'irf.width'},
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['m1'],
                'shape': {
                    'osc1': 'sh1',
                    's1': 'sh2',
                    's2': 'sh3',
                },
                'irf': 'irf1',
            }
        }
    })

    model = DOASModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1', 's2'],
                'parameters': ['j.1', 'j.0']
            },
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s2", "s1"): 'kinetic.1',
                ("s2", "s2"): 'kinetic.2',
            }}
        },
        'oscillation': {
            'osc1': {'frequency': 'osc.freq', 'rate': 'osc.rate'}
        },
        'megacomplex': {
            'm1': {
                'k_matrix': ['k1'],
                'oscillation': ['osc1'],
            }
        },
        'irf': {
            'irf1': {'type': 'gaussian', 'center': 'irf.center', 'width': 'irf.width'},
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['m1'],
                'irf': 'irf1',
            }
        }
    })

    wanted_parameter = ParameterGroup.from_dict({
        'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
            ['0', 0, {'vary': False, 'non-negative': False}],
        ],
        'kinetic': [
            ["1", 0.2],
            ["2", 0.01],
        ],
        'osc': [
            ['freq', 25],
            ['rate', 0.1],
        ],
        'shape': {'amps': [0.07, 2, 4], 'locs': [5, 2, 8], 'width': [4, 2, 3]},
        'irf': [['center', 0.3], ['width', 0.1]],
    })

    parameter = ParameterGroup.from_dict({
        'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
            ['0', 0, {'vary': False, 'non-negative': False}],
        ],
        'kinetic': [
            ["1", 0.2],
            ["2", 0.01],
        ],
        'osc': [
            ['freq', 25],
            ['rate', 0.1],
        ],
        'irf': [['center', 0.3], ['width', 0.1]],
    })

    time = np.arange(-1, 5, 0.01)
    spectral = np.arange(0, 10)
    axis = {'time': time, 'spectral': spectral}

    wanted_clp = ['osc1_cos', 'osc1_sin', 's1', 's2']
    wanted_shape = (600, 4)


@pytest.mark.parametrize("suite", [
    OneOscillation,
    OneOscillationWithIrf,
    OneOscillationWithSequentialModel,
])
def test_doas_model(suite):

    print(suite.sim_model.validate())
    assert suite.sim_model.valid()

    print(suite.model.validate())
    assert suite.model.valid()

    print(suite.sim_model.validate(suite.wanted_parameter))
    assert suite.sim_model.valid(suite.wanted_parameter)

    print(suite.model.validate(suite.parameter))
    assert suite.model.valid(suite.parameter)

    dataset = suite.sim_model.dataset['dataset1'].fill(suite.sim_model, suite.wanted_parameter)
    data = xr.DataArray(suite.axis['time'], coords=[('time', suite.axis['time'])])

    clp, matrix = calculate_doas_matrix(dataset, suite.axis['time'], 0)

    print(matrix.shape)
    assert matrix.shape == suite.wanted_shape

    print(clp)
    assert clp == suite.wanted_clp

    dataset = suite.sim_model.simulate('dataset1', suite.wanted_parameter,
                                       suite.axis)
    print(dataset)

    assert dataset.data.shape == \
        (suite.axis['time'].size, suite.axis['spectral'].size)

    print(suite.parameter)
    print(suite.wanted_parameter)
    data = {'dataset1': dataset}

    result = suite.model.optimize(suite.parameter, data, max_nfev=50)
    print(result.optimized_parameter)

    for label, param in result.optimized_parameter.all():
        assert np.allclose(param.value, suite.wanted_parameter.get(label).value,
                           rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset['time'], resultdata['time'])
    assert np.array_equal(dataset['spectral'], resultdata['spectral'])
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data)

    assert 'dampened_oscillation_cos' in resultdata
    assert 'dampened_oscillation_sin' in resultdata
    assert 'dampened_oscillation_associated_spectra' in resultdata
    assert 'dampened_oscillation_phase' in resultdata



"""The result class for global analysis."""

import typing
import os

import numpy as np
import xarray as xr

import glotaran  # noqa F01
from glotaran.parameter import ParameterGroup

from .scheme import Scheme


class Result:

    def __init__(self,
                 scheme: Scheme,
                 data: typing.Dict[str, xr.Dataset],
                 parameter: ParameterGroup,
                 nfev: int,
                 nvars: int,
                 ndata: int,
                 nfree: int,
                 chisqr: float,
                 red_chisqr: float,
                 var_names: typing.List[str],
                 covar: np.ndarray,
                 ):
        """The result of a global analysis.

        Parameters
        ----------
        model :
            A subclass of :class:`glotaran.model.Model`
        data :
            A dictonary containing all datasets with their labels as keys.
        initital_parameter : glotaran.parameter.ParameterGroup
            The initital fit parameter,
        nnls :
            (default = False)
            If `True` non-linear least squaes optimizing is used instead of variable projection.
        atol :
            (default = 0)
            The tolerance for grouping datasets along the global axis.
        """
        self._scheme = scheme
        self._data = data
        self._optimized_parameter = parameter
        self._nfev = nfev
        self._nvars = nvars
        self._ndata = ndata,
        self._nfree = nfree
        self._chisqr = chisqr
        self._red_chisqr = red_chisqr
        self._var_names = var_names
        self._covar = covar

    @property
    def scheme(self) -> Scheme:
        """The scheme for analysis."""
        return self._scheme

    @property
    def model(self) -> typing.Type['glotaran.model.Model']:
        """The model for analysis."""
        return self._scheme.model

    @property
    def nnls(self) -> bool:
        """If `True` non-linear least squaes optimizing is used instead of variable
        projection."""
        return self._scheme.nnls

    @property
    def data(self) -> typing.Dict[str, xr.Dataset]:
        """The resulting data as a dictionary of :xarraydoc:`Dataset`.

        Notes
        -----
        The actual content of the data depends on the actual model and can be found in the
        documentation for the model.
        """
        return self._data

    @property
    def nfev(self) -> int:
        """The number of function evaluations."""
        return self._nfev

    @property
    def nvars(self) -> int:
        """Number of variables in optimization :math:`N_{vars}`"""
        return self._nvars

    @property
    def ndata(self) -> int:
        """Number of data points :math:`N`."""
        return self._ndata

    @property
    def nfree(self) -> int:
        """Degrees of freedom in optimization :math:`N - N_{vars}`."""
        return self._nfree

    @property
    def chisqr(self) -> float:
        """The chi-square of the optimization
        :math:`\chi^2 = \sum_i^N [{Residual}_i]^2`.""" # noqa w605
        return self._chisqr

    @property
    def red_chisqr(self) -> float:
        """The reduced chi-square of the optimization
        :math:`\chi^2_{red}= {\chi^2} / {(N - N_{vars})}`.
        """ # noqa w605
        return self._red_chisqr

    @property
    def root_mean_square_error(self) -> float:
        """
        The root mean square error the optimization
        :math:`rms = \sqrt{\chi^2_{red}}`
        """ # noqa w605
        return np.sqrt(self.red_chisqr)

    @property
    def var_names(self) -> typing.List[str]:
        """Ordered list of variable parameter names used in optimization, and
        useful for understanding the values in :attr:`covar`."""
        return [n.replace('_', '.') for n in self._var_names]

    @property
    def covar(self) -> np.ndarray:
        """Covariance matrix from minimization, with rows and columns
        corresponding to :attr:`var_names`."""
        return self._covar

    @property
    def optimized_parameter(self) -> ParameterGroup:
        """The optimized parameters."""
        return self._optimized_parameter

    @property
    def initial_parameter(self) -> ParameterGroup:
        """The initital fit parameter"""
        return self._scheme.parameter

    def get_dataset(self, dataset_label: str) -> xr.Dataset:
        """Returns the result dataset for the given dataset label.

        Parameters
        ----------
        dataset_label :
            The label of the dataset.
        """
        try:
            return self.data[dataset_label]
        except KeyError:
            raise Exception(f"Unknown dataset '{dataset_label}'")

    def save(self,  path: str) -> typing.List[str]:
        """Saves the result to given folder.

        Returns a list with paths of all saved items.

        The following files are saved:

        * `result.md`: The result with the model formatted as markdown text.
        * `optimized_parameter.csv`: The optimized parameter as csv file.
        * `{dataset_label}.nc`: The result data for each dataset as NetCDF file.

        Parameters
        ----------
        path :
            The path to the folder in which to save the result.
        """
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            raise Exception(f"The path '{path}' is not a directory.")

        paths = []

        md_path = os.path.join(path, 'result.md')
        with open(md_path, 'w') as f:
            f.write(self.markdown())
        paths.append(md_path)

        csv_path = os.path.join(path, 'optimized_parameter.csv')
        self.optimized_parameter.to_csv(csv_path)
        paths.append(csv_path)

        for label, data in self.data.items():
            nc_path = os.path.join(path, f"{label}.nc")
            data.to_netcdf(nc_path, engine='netcdf4')
            paths.append(nc_path)

        return paths

    def markdown(self, with_model=True) -> str:
        """Formats the model as a markdown text.

        Parameters
        ----------
        with_model :
            If `True`, the model will be printed together with the initial and optimized parameter.
        """

        ll = 32
        lr = 13

        string = "Optimization Result".ljust(ll-1)
        string += "|"
        string += "|".rjust(lr)
        string += "\n"
        string += "|".rjust(ll, "-")
        string += "|".rjust(lr, "-")
        string += "\n"

        string += "Number of residual evaluation |".rjust(ll)
        string += f"{self.nfev} |".rjust(lr)
        string += "\n"
        string += "Number of variables |".rjust(ll)
        string += f"{self.nvars} |".rjust(lr)
        string += "\n"
        string += "Number of datapoints |".rjust(ll)
        string += f"{self.ndata} |".rjust(lr)
        string += "\n"
        string += "Degrees of freedom |".rjust(ll)
        string += f"{self.nfree} |".rjust(lr)
        string += "\n"
        string += "Chi Square |".rjust(ll)
        string += f"{self.chisqr:.2e} |".rjust(lr)
        string += "\n"
        string += "Reduced Chi Square |".rjust(ll)
        string += f"{self.red_chisqr:.2e} |".rjust(lr)
        string += "\n"
        string += "Root Mean Square Error |".rjust(ll)
        string += f"{self.root_mean_square_error:.2e} |".rjust(lr)
        string += "\n"

        if with_model:
            string += "\n\n" + self.model.markdown(parameter=self.optimized_parameter,
                                                   initial=self.initial_parameter)

        return string

    def __str__(self):
        return self.markdown(with_model=False)



"""Glotaran Model Package

This package contains the Glotaran's base model object, the model decorators and
common model items.
"""

from . import (
    dataset_descriptor,
    model,
    model_decorator,
    model_attribute,
)

# Dataset

DatasetDescriptor = dataset_descriptor.DatasetDescriptor

# BaseModel

Model = model.Model

# Decorators

model = model_decorator.model
model_attribute_typed = model_attribute.model_attribute_typed
model_attribute = model_attribute.model_attribute



import pathlib
import xarray as xr

known_reading_formats = {
}


def read_data_file(filename: str, fmt: str = None) -> xr.Dataset:
    path = pathlib.Path(filename)

    if fmt is None:
        fmt = path.suffix[1:] if path.suffix != '' else 'nc'

    if fmt not in known_reading_formats:
        raise Exception(
            f"Unknown file format '{fmt}'."
            f"Supported formats are {list(known_reading_formats.keys())}")

    return known_reading_formats[fmt].read(filename)


class Reader:

    def __init__(self, extension, name, reader):
        self.extension = extension
        self.name = name if name else f" '.{extension}' format"
        self.read = reader


def file_reader(extension: str = None, name: str = None):

    def decorator(reader):
        known_reading_formats[extension] = Reader(extension, name, reader)
        return reader

    return decorator


file_reader(extension='nc', name='netCDF4')(xr.open_dataset)



import numpy as np

from glotaran.parameter import ParameterGroup


def test_param_array():
    params = """
    - 5
    - 4
    - 3
    - 2
    - 1
    """

    params = ParameterGroup.from_yaml(params)

    assert len(list(params.all())) == 5

    assert [p.label for _, p in params.all()] == [f"{i}" for i in range(1, 6)]
    assert [p.value for _, p in params.all()] == list(range(1, 6))[::-1]


def test_param_label():
    params = """
    - ["5", 1]
    - ["4", 2]
    - ["3", 3]
    """

    params = ParameterGroup.from_yaml(params)

    assert len(list(params.all())) == 3
    assert [p.label for _, p in params.all()] == [f"{i}" for i in range(5, 2, -1)]
    assert [p.value for _, p in params.all()] == list(range(1, 4))


def test_param_options():
    params = """
    - ["5", 1, {non-negative: false, min: -1, max: 1, vary: false}]
    - ["6", 4e2, {non-negative: true, min: -7e2, max: 8e2, vary: true}]
    - ["7", 2e4]
    """

    params = ParameterGroup.from_yaml(params)

    assert params.get("5").value == 1.0
    assert not params.get("5").non_neg
    assert params.get("5").min == -1
    assert params.get("5").max == 1
    assert not params.get("5").vary

    assert params.get("6").value == 4e2
    assert params.get("6").non_neg
    assert params.get("6").min == -7e2
    assert params.get("6").max == 8e2
    assert params.get("6").vary

    assert params.get("7").value == 2e4
    assert not params.get("7").non_neg
    assert params.get("7").min == float('-inf')
    assert params.get("7").max == float('inf')
    assert params.get("7").vary


def test_param_block_options():
    params = """
    block:
        - 1.0
        - [3.4, {vary: true}]
        - {vary: false}
    """

    params = ParameterGroup.from_yaml(params)
    assert not params.get("block.1").vary
    assert params.get("block.2").vary


def test_nested_param_list():
    params = """
    kinetic:
        - 3
        - 4
        - 5
    j:
        - 7
        - 8
    """

    params = ParameterGroup.from_yaml(params)

    assert len(list(params.all())) == 5
    group = params['kinetic']
    assert len(list(group.all())) == 3
    assert [p.label for _, p in group.all()] == [f"{i}" for i in range(1, 4)]
    assert [p.value for _, p in group.all()] == list(range(3, 6))
    group = params['j']
    assert len(list(group.all())) == 2
    assert [p.label for _, p in group.all()] == [f"{i}" for i in range(1, 3)]
    assert [p.value for _, p in group.all()] == list(range(7, 9))


def test_nested_param_group():
    params = """
    kinetic:
        j:
            - 7
            - 8
            - 9
    """

    params = ParameterGroup.from_yaml(params)
    assert len(list(params.all())) == 3
    group = params['kinetic']
    assert len(list(group.all())) == 3
    group = group['j']
    assert len(list(group.all())) == 3
    assert [p.label for _, p in group.all()] == [f"{i}" for i in range(1, 4)]
    assert [p.value for _, p in group.all()] == list(range(7, 10))


def test_non_negative():

    params = """
    - ["neg", -1]
    - ["negmax", -1, {max=0}]
    - ["nonneg1", 1, {non-negative: True}]
    - ["nonneg2", 2, {non-negative: True}]
    - ["nonnegmin", 6, {non-negative: True, min: 2}]
    """
    params = ParameterGroup.from_yaml(params)
    result = ParameterGroup.from_parameter_dict(params.as_parameter_dict())
    print(params)
    params.as_parameter_dict().pretty_print()
    print(result)

    for label, p in params.all():
        print(label)
        r = result.get(label)
        assert r.non_neg == p.non_neg
        assert np.allclose(r.value, p.value)
        assert np.allclose(r.min, p.min)
        assert np.allclose(r.max, p.max)



""" Glotaran Kinetic Matrix """

import numpy as np
import numba as nb

from .irf import IrfMultiGaussian


def kinetic_image_matrix(
        dataset_descriptor=None, axis=None, index=None, irf=None):
    return kinetic_matrix(
        dataset_descriptor, axis, index, irf, kinetic_image_matrix_implementation)


def kinetic_matrix(
        dataset_descriptor=None, axis=None, index=None, irf=None, matrix_implementation=None):

    compartments = None
    matrix = None
    k_matrices = dataset_descriptor.get_k_matrices()

    if len(k_matrices) == 0:
        return (None, None)

    if dataset_descriptor.initial_concentration is None:
        raise Exception(
            f'No initial concentration specified in dataset "{dataset_descriptor.label}"')
    initial_concentration = dataset_descriptor.initial_concentration.normalized(dataset_descriptor)

    for k_matrix in k_matrices:

        if k_matrix is None:
            continue

        (this_compartments, this_matrix) = _calculate_for_k_matrix(
            dataset_descriptor,
            axis,
            index,
            k_matrix,
            initial_concentration,
            irf,
            matrix_implementation
        )

        if matrix is None:
            compartments = this_compartments
            matrix = this_matrix
        else:
            new_compartments = \
                    compartments + [c for c in this_compartments if c not in compartments]
            new_matrix = np.zeros((matrix.shape[0], len(new_compartments)), dtype=np.float64)
            for i, comp in enumerate(new_compartments):
                if comp in compartments:
                    new_matrix[:, i] += matrix[:, compartments.index(comp)]
                if comp in this_compartments:
                    new_matrix[:, i] += this_matrix[:, this_compartments.index(comp)]
            compartments = new_compartments
            matrix = new_matrix

    if dataset_descriptor.baseline:
        baseline_compartment = f'{dataset_descriptor.label}_baseline'
        baseline = np.ones((axis.size, 1), dtype=np.float64)
        if matrix is None:
            compartments = [baseline_compartment]
            matrix = baseline
        else:
            compartments.append(baseline_compartment)
            matrix = np.concatenate((matrix, baseline), axis=1)

    return (compartments, matrix)


def _calculate_for_k_matrix(dataset_descriptor, axis, index, k_matrix,
                            initial_concentration, irf, matrix_implementation):

    # we might have more compartments in the model then in the k matrix
    compartments = [comp for comp in initial_concentration.compartments
                    if comp in k_matrix.involved_compartments()]

    # the rates are the eigenvalues of the k matrix
    rates = k_matrix.rates(initial_concentration)

    # init the matrix
    size = (axis.size, rates.size)
    matrix = np.zeros(size, dtype=np.float64)

    matrix_implementation(matrix, rates, axis, index, dataset_descriptor, irf)

    if not np.all(np.isfinite(matrix)):
        raise Exception(f"Non-finite concentrations for K-Matrix '{k_matrix.label}':\n"
                        "{k_matrix.matrix_as_markdown}")

    # apply A matrix
    matrix = matrix @ k_matrix.a_matrix(initial_concentration)

    # done
    return (compartments, matrix)


def kinetic_image_matrix_implementation(
        matrix, rates, axis, index, dataset_descriptor, measured_irf):

    if isinstance(dataset_descriptor.irf, IrfMultiGaussian):

        center, width, irf_scale, backsweep, backsweep_period = \
            dataset_descriptor.irf.parameter(index)

        for i in range(len(center)):
            calculate_kinetic_matrix_gaussian_irf(
                matrix, rates, axis, center[i], width[i],
                irf_scale[i], backsweep, backsweep_period,)
        matrix /= np.sum(irf_scale)

    else:
        calculate_kinetic_matrix_no_irf(matrix, rates, axis)


@nb.jit(nopython=True, parallel=True)
def calculate_kinetic_matrix_no_irf(matrix, rates, times):
    for n_r in nb.prange(rates.size):
        r_n = rates[n_r]
        for n_t in range(times.size):
            t_n = times[n_t]
            matrix[n_t, n_r] += np.exp(r_n * t_n)


sqrt2 = np.sqrt(2)
@nb.jit(nopython=True, parallel=True)
def calculate_kinetic_matrix_gaussian_irf(
        matrix, rates, times, center, width, scale, backsweep, backsweep_period):
    """Calculates a kinetic matrix with a gaussian irf."""
    for n_r in nb.prange(rates.size):
        r_n = -rates[n_r]
        alpha = (r_n * width) / sqrt2
        for n_t in nb.prange(times.size):
            t_n = times[n_t]
            beta = (t_n - center) / (width * sqrt2)
            thresh = beta - alpha
            if thresh < -1:
                matrix[n_t, n_r] += scale * .5 * erfcx(-thresh) * np.exp(-beta * beta)
            else:
                matrix[n_t, n_r] += \
                        scale * .5 * (1 + erf(thresh)) * np.exp(alpha * (alpha - 2 * beta))
            if backsweep:
                x1 = np.exp(-r_n * (t_n - center + backsweep_period))
                x2 = np.exp(-r_n * ((backsweep_period / 2) - (t_n - center)))
                x3 = np.exp(-r_n * backsweep_period)
                matrix[n_t, n_r] += scale * (x1 + x2) / (1 - x3)


# This is a work around to use scipy.special function with numba
from numba.extending import get_cython_function_address # noqa
import ctypes # noqa

_dble = ctypes.c_double

functype = ctypes.CFUNCTYPE(_dble, _dble)

erf_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erf")
erfcx_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erfcx")

erf = functype(erf_addr)
erfcx = functype(erfcx_addr)



import functools
import pathlib
import typing
import yaml
import xarray as xr
import numpy as np

import glotaran
from glotaran.parameter import ParameterGroup


def _not_none(f):

    @functools.wraps(f)
    def decorator(self, value):
        if value is None:
            raise ValueError(f"{f.__name__} cannot be None")
        f(self, value)


class Scheme:

    def __init__(self,
                 model: 'glotaran.model.Model' = None,
                 parameter: ParameterGroup = None,
                 data: typing.Dict[str, typing.Union[xr.DataArray, xr.Dataset]] = None,
                 group_tolerance: float = 0.0,
                 nnls: bool = False,
                 nfev: int = None,
                 ):

        self.model = model
        self.parameter = parameter
        self.data = data
        self.group_tolerance = group_tolerance
        self.nnls = nnls
        self.nfev = nfev

    @classmethod
    def from_yml_file(cls, filename: str) -> 'Scheme':

        try:
            with open(filename) as f:
                try:
                    scheme = yaml.load(f, Loader=yaml.FullLoader)
                except Exception as e:
                    raise Exception(f"Error parsing scheme: {e}")
        except Exception as e:
            raise Exception(f"Error opening scheme: {e}")

        if 'model' not in scheme:
            raise Exception('Model file not specified.')

        try:
            model = glotaran.read_model_from_yml_file(scheme['model'])
        except Exception as e:
            raise Exception(f"Error loading model: {e}")

        if 'parameter' not in scheme:
            raise Exception('Parameter file not specified.')

        path = scheme['parameter']
        fmt = scheme.get('parameter_format', None)
        try:
            parameter = glotaran.parameter.ParameterGroup.from_file(path, fmt)
        except Exception as e:
            raise Exception(f"Error loading parameter: {e}")

        if 'data' not in scheme:
            raise Exception('No data specified.')

        data = {}
        for label, path in scheme['data'].items():
            path = pathlib.Path(path)

            fmt = path.suffix[1:] if path.suffix != '' else 'nc'
            if 'data_format' in scheme:
                fmt = scheme['data_format']

            try:
                data[label] = glotaran.io.read_data_file(path, fmt=fmt)
            except Exception as e:
                raise Exception(f"Error loading dataset '{label}': {e}")

        nnls = scheme.get('nnls', False)
        nfev = scheme.get('nfev', None)
        group_tolerance = scheme.get('group_tolerance', 0.0)
        return cls(model=model, parameter=parameter, data=data,
                   nnls=nnls, nfev=nfev, group_tolerance=group_tolerance)

    @property
    def model(self) -> 'glotaran.model.Model':
        return self._model

    @_not_none
    @model.setter
    def model(self, model: 'glotaran.model.Model'):
        self._model = model

    @property
    def parameter(self) -> ParameterGroup:
        return self._parameter

    @_not_none
    @parameter.setter
    def parameter(self, parameter: ParameterGroup):
        self._parameter = parameter

    @property
    def data(self) -> typing.Dict[str, typing.Union[xr.DataArray, xr.Dataset]]:
        return self._data

    @_not_none
    @data.setter
    def data(self, data: typing.Dict[str, typing.Union[xr.DataArray, xr.Dataset]]):
        self._data = data

    @property
    def nnls(self) -> bool:
        return self._nnls

    @_not_none
    @nnls.setter
    def nnls(self, nnls: bool):
        self._nnls = nnls

    @property
    def nfev(self) -> int:
        return self._nfev

    @nfev.setter
    def nfev(self, nfev: int):
        self._nfev = nfev

    @property
    def group_tolerance(self) -> float:
        return self._group_tolerance

    @group_tolerance.setter
    def group_tolerance(self, group_tolerance: float):
        self._group_tolerance = group_tolerance

    def problem_list(self) -> typing.List[str]:
        """Returns a list with all problems in the model and missing parameters."""
        return self.model.problem_list(self.parameter)

    def validate(self) -> str:
        """Returns a string listing all problems in the model and missing parameters."""
        return self.model.validate(self.parameter)

    def valid(self, parameter: ParameterGroup = None) -> bool:
        """Returns `True` if there are no problems with the model or the parameters,
        else `False`."""
        return self.model.valid(parameter)

    def prepared_data(self) -> typing.Dict[str, xr.Dataset]:
        data = {}
        for label, dataset in self.data.items():
            if self.model.matrix_dimension not in dataset.dims:
                raise Exception("Missing coordinates for dimension "
                                f"'{self.model.matrix_dimension}' in data for dataset "
                                f"'{label}'")
            if self.model.global_dimension not in dataset.dims:
                raise Exception("Missing coordinates for dimension "
                                f"'{self.model.global_dimension}' in data for dataset "
                                f"'{label}'")
            if isinstance(dataset, xr.DataArray):
                dataset = dataset.to_dataset(name="data")

            if 'weight' in dataset and 'weighted_data' not in dataset:
                dataset['weighted_data'] = np.multiply(dataset.data, dataset.weight)

            # This protects transposing when getting data with svd in it

            if 'data_singular_values' in dataset:
                if dataset.coords['right_singular_value_index'].size != \
                  dataset.coords[self.model.global_dimension].size:
                    dataset = dataset.rename(
                        right_singular_value_index='right_singular_value_indexTMP')
                    dataset = dataset.rename(
                        left_singular_value_index='right_singular_value_index')
                    dataset = dataset.rename(
                        right_singular_value_indexTMP='left_singular_value_index')
                    dataset = dataset.rename(
                        right_singular_vectors='right_singular_value_vectorsTMP')
                    dataset = dataset.rename(
                        left_singular_value_vectors='right_singular_value_vectors')
                    dataset = dataset.rename(
                        right_singular_value_vectorsTMP='left_singular_value_vectors')
            new_dims = [self.model.matrix_dimension, self.model.global_dimension]
            new_dims += [dim for dim in dataset.dims
                         if dim != self.model.matrix_dimension
                         and dim != self.model.global_dimension]
            data[label] = dataset.transpose(*new_dims)
        return data

    def markdown(self):
        s = self.model.markdown(parameter=self.parameter)

        s += "\n\n"
        s += "__Scheme__\n\n"

        s += f"* *nnls*: {self.nnls}\n"
        s += f"* *nfev*: {self.nfev}\n"
        s += f"* *group_tolerance*: {self.group_tolerance}\n"

        return s



import click

from glotaran.io.reader import known_reading_formats
from glotaran.parse.register import known_model_names


def plugin_list_cmd():
    """Prints a list of intalled plugins."""

    output = """
Installed Glotaran Plugins:

    Models:
    """
    output += "\n"

    for name in known_model_names():
        output += f"    * {name}\n"

    output += "\nFile Formats\n\n"

    for reader in known_reading_formats.values():
        output += f"    * .{reader.extension} : {reader.name}\n"

    click.echo(output)



import typing
import numpy as np
import xarray as xr

import glotaran
from glotaran.parameter import ParameterGroup

from .irf import IrfMultiGaussian


def finalize_kinetic_image_result(
    model: 'glotaran.models.spectral_temporal.KineticModel',
    global_indices: typing.List[typing.List[object]],
    reduced_clp_labels: typing.Union[typing.Dict[str, typing.List[str]], np.ndarray],
    reduced_clps: typing.Union[typing.Dict[str, np.ndarray], np.ndarray],
    parameter: ParameterGroup, data: typing.Dict[str, xr.Dataset],
):

    for label in model.dataset:
        dataset = data[label]

        dataset_descriptor = model.dataset[label].fill(model, parameter)

        if not dataset_descriptor.get_k_matrices():
            continue

        retrieve_species_assocatiated_data(model, dataset, dataset_descriptor, "images")
        retrieve_decay_assocatiated_data(model, dataset, dataset_descriptor, "images")

        if dataset_descriptor.baseline:
            dataset['baseline'] = dataset.clp.sel(clp_label=f"{dataset_descriptor.label}_baseline")

        retrieve_irf(model, dataset, dataset_descriptor, "images")


def retrieve_species_assocatiated_data(model, dataset, dataset_descriptor, name):
    compartments = dataset_descriptor.initial_concentration.compartments

    compartments = [c for c in compartments if c in dataset_descriptor.compartments()]

    dataset.coords['species'] = compartments
    dataset[f'species_associated_{name}'] = \
        ((model.global_dimension, 'species',),
         dataset.clp.sel(clp_label=compartments))

    if len(dataset.matrix.shape) == 3:
        #  index dependent
        dataset['species_concentration'] = (
            (model.global_dimension, model.matrix_dimension, 'species',),
            dataset.matrix.sel(clp_label=compartments).values)
    else:
        #  index independent
        dataset['species_concentration'] = (
            (model.matrix_dimension, 'species',),
            dataset.matrix.sel(clp_label=compartments).values)


def retrieve_decay_assocatiated_data(model, dataset, dataset_descriptor, name):
    # get_das
    all_das = []
    all_a_matrix = []
    all_k_matrix = []
    all_das_labels = []
    for megacomplex in dataset_descriptor.megacomplex:

        k_matrix = megacomplex.full_k_matrix()
        if k_matrix is None:
            continue

        compartments = dataset_descriptor.initial_concentration.compartments
        compartments = [c for c in compartments if c in k_matrix.involved_compartments()]

        matrix = k_matrix.full(compartments)
        a_matrix = k_matrix.a_matrix(dataset_descriptor.initial_concentration)
        rates = k_matrix.rates(dataset_descriptor.initial_concentration)
        lifetimes = 1/rates

        das = dataset[f'species_associated_{name}'].sel(species=compartments).values @ a_matrix.T

        component_coords = {
            'rate': ('component', rates),
            'lifetime': ('component', lifetimes)}

        das_coords = component_coords.copy()
        das_coords[model.global_dimension] = dataset.coords[model.global_dimension]
        all_das_labels.append(megacomplex.label)
        all_das.append(
            xr.DataArray(
                das, dims=(model.global_dimension, 'component'),
                coords=das_coords))
        a_matrix_coords = component_coords.copy()
        a_matrix_coords['species'] = compartments
        all_a_matrix.append(
            xr.DataArray(a_matrix, coords=a_matrix_coords, dims=('component', 'species')))
        all_k_matrix.append(
            xr.DataArray(matrix, coords=[('to_species', compartments),
                                         ('from_species', compartments)]))

    if all_das:
        if len(all_das) == 1:
            dataset[f'decay_associated_{name}'] = all_das[0]
            dataset['a_matrix'] = all_a_matrix[0]
            dataset['k_matrix'] = all_k_matrix[0]
        else:
            for i, das_label in enumerate(all_das_labels):
                dataset[f'decay_associated_{name}_{das_label}'] = \
                        all_das[i].rename(component=f"component_{das_label}")
                dataset[f'a_matrix_{das_label}'] = all_a_matrix[i] \
                    .rename(component=f"component_{das_label}")
                dataset[f'k_matrix_{das_label}'] = all_k_matrix[i]


def retrieve_irf(model, dataset, dataset_descriptor, name):

    irf = dataset_descriptor.irf

    if isinstance(irf, IrfMultiGaussian):
        index = dataset.coords[model.global_dimension][0].values
        dataset['irf'] = ((model.matrix_dimension), irf.calculate(index, dataset.coords['time']))



"""Glotaran DOAS Matrix"""

import numba as nb
import numpy as np
from scipy.special import erf

from glotaran.builtin.models.kinetic_image.kinetic_image_matrix \
    import kinetic_image_matrix
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian


def calculate_doas_matrix(dataset_descriptor=None, axis=None, index=None, irf=None):

    oscillations = _collect_oscillations(dataset_descriptor)
    matrix = np.zeros((axis.size, len(2 * oscillations)), dtype=np.float64)
    labels = [o.label for o in oscillations]
    clp = [
        lbl for o in zip([f'{l}_cos' for l in labels], [f'{l}_sin' for l in labels]) for lbl in o
    ]

    delta = np.abs(axis[1:] - axis[:-1])
    delta_min = delta[np.argmin(delta)]
    frequency_max = 1 / (2 * 0.03 * delta_min)
    frequencies = np.asarray([o.frequency * 0.03 * 2 * np.pi for o in oscillations])
    frequencies[frequencies >= frequency_max] = \
        np.mod(frequencies[frequencies >= frequency_max], frequency_max)
    rates = np.asarray([o.rate for o in oscillations])

    if dataset_descriptor.irf is None:
        calculate_doas_matrix_no_irf(matrix, frequencies, rates, axis)
    elif isinstance(dataset_descriptor.irf, IrfMultiGaussian):
        centers, widths, scales, _, _ = dataset_descriptor.irf.parameter(index)
        calculate_doas_matrix_gaussian_irf(
            matrix, frequencies, rates, axis, centers, widths, scales)

    kinetic_clp, kinetic_matrix = kinetic_image_matrix(dataset_descriptor, axis, index, irf)
    if kinetic_matrix is not None:
        clp = clp + kinetic_clp
        matrix = np.concatenate((matrix, kinetic_matrix), axis=1)
    return (clp, matrix)


@nb.jit(nopython=True, parallel=True)
def calculate_doas_matrix_no_irf(matrix, frequencies, rates, axis):

    idx = 0
    for frequency, rate in zip(frequencies, rates):
        osc = np.exp(-rate * axis - 1j * frequency * axis)
        matrix[:, idx] = osc.real
        matrix[:, idx + 1] = osc.imag
        idx += 2


def calculate_doas_matrix_gaussian_irf(matrix, frequencies, rates, axis, centers, widths, scales):

    idx = 0
    for frequency, rate in zip(frequencies, rates):
        osc = np.zeros_like(axis, dtype=np.complex64)
        for i in range(len(centers)):
            shifted_axis = axis - centers[i]
            d = widths[i]**2
            k = (rate + 1j * frequency)

            a = (-1 * shifted_axis + 0.5 * d * k) * k
            a = np.minimum(a, 709)
            a = np.exp(a)
            b = 1 + erf((shifted_axis - d * k) / (np.sqrt(2) * widths[i]))
            if not np.all(np.isfinite(a)):
                raise Exception("Non numeric values in term 'a' for oscillation with frequency.")
            if not np.all(np.isfinite(b)):
                idx = np.where(np.logical_not(np.isfinite(b)))[0]
                raise Exception("Non numeric values in term 'b' for oscillation.")
            osc = a * b * scales[i]
        osc /= np.sum(scales)
        matrix[:, idx] = osc.real
        matrix[:, idx + 1] = osc.imag
        idx += 2


def _collect_oscillations(dataset):
    return [osc for cmplx in dataset.megacomplex for osc in cmplx.oscillation]



"""
Glotarans module to read files
"""
import warnings

import numpy as np
import xarray as xr

from glotaran.io.reader import file_reader
from glotaran.io.prepare_dataset import prepare_time_trace_dataset

from .sdt_file import SdtFile


@file_reader(extension='sdt', name='Becker & Hickel SDT file format.')
def read_sdt(file_path: str,
             index: np.ndarray = None,
             flim: bool = False,
             dataset_index: int = None,
             swap_axis: bool = False,
             orig_time_axis_index: int = 2,
             ) -> xr.Dataset:
    """
    Reads a `*.sdt` file and returns a pd.DataFrame (`return_dataframe==True`), a
    SpectralTemporalDataset (`type_of_data=='st'`) or a FLIMDataset (`type_of_data=='flim'`).

    Parameters
    ----------
    file_path: str
        Path to the sdt file which should be read.

    index: list, np.ndarray
        This is only needed if `type_of_data=="st"`, since `*.sdt` files,
        which only contain spectral temporal data, lack the spectral information.
        Thus for the spectral axis data need to be given by the user.

    flim:
        Set true if reading a result from a FLIM measurement.

    dataset_index: int: default 0
        If the `*.sdt` file contains multiple datasets the index will used
        to select the wanted one

    swap_axis: bool, default False
        Flag to switch a wavelength explicit `input_df` to time explicit `input_df`,
        before generating the SpectralTemporalDataset.

    orig_time_axis_index: int
        Index of the axis which corresponds to the time axis.
        I.e. for data of shape (64, 64, 256), which are a 64x64 pixel map
        with 256 time steps, orig_time_axis_index=2.

    Raises
    ______
    IndexError:
        If the length of the index array is incompatible with the data.
    """
    sdt_parser = SdtFile(file_path)
    if not dataset_index:
        # looking at the source code of SdtFile, times and data
        # always have the same len, so only one needs to be checked
        nr_of_datasets = len(sdt_parser.times)
        if nr_of_datasets > 1:
            warnings.warn(UserWarning(
                f"The file '{file_path}' contains {nr_of_datasets} Datasets.\n "
                f"By default only the first Dataset will be read. "
                f"If you only need the first Dataset and want get rid of "
                f"this warning you can set dataset_index=0.")
            )
        dataset_index = 0
    times = sdt_parser.times[dataset_index]
    data = sdt_parser.data[dataset_index]

    if index and len(index) is not data.shape[0]:
        raise IndexError(f"The Dataset contains {data.shape[0]} measurements, but the "
                         f"indices supplied are {len(index)}.")
    elif not index and not flim:
        warnings.warn(UserWarning(
            f"There was no `index` provided."
            f"That for the indices will be a entry count(integers)."
            f"To prevent this warning from being shown, provide "
            f"a list of indices, with len(index)={data.shape[0]}")
            )

    if flim:

        if orig_time_axis_index != 2:
            np.swapaxes(data, 2, orig_time_axis_index)

        full_data = xr.DataArray(data, coords={'time': times}, dims=['x', 'y', 'time'])
        data = full_data.stack(pixel=('x', 'y')).to_dataset(name='data')
        data['full_data'] = full_data.rename({'x': 'pixel_x', 'y': 'pixel_y'})
        data['data_intensity_map'] = \
            data.data.groupby('pixel').sum().unstack().rename({'x': 'pixel_x', 'y': 'pixel_y'})
    else:
        if swap_axis:
            data = data.T
        if not index:
            index = np.array(range(data[0]))
        data = xr.DataArray(data.T, coords=[('time', times), ('spectral', index)])
        data = prepare_time_trace_dataset(data)
    return data



import click

from . import util


@util.signature_analysis
def print_cmd(parameter: str, model: str, scheme: str):
    """Parses scheme, a model or a parameter file and prints the result as a Markdown formatted
    string."""

    if scheme is not None:
        scheme = util.load_scheme_file(scheme, verbose=False)
        click.echo(scheme.markdown())
        return
    if model is not None:
        model = util.load_model_file(model, verbose=False)

    if parameter is not None:
        parameter = util.load_parameter_file(parameter, verbose=False)

    if model:
        click.echo(model.markdown(parameter=parameter))
    elif parameter:
        click.echo(parameter.markdown())
    else:
        click.echo("Nothing to print, please type 'glotaran print --help' for more info.")



""" Gloataran DOAS Model """

from glotaran.model import model
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_dataset_descriptor import \
    KineticSpectrumDatasetDescriptor
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model import (
    KineticSpectrumModel,
    apply_kinetic_model_constraints,
    grouped,
    index_dependend,
)
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import apply_spectral_penalties

from .doas_result import finalize_doas_data
from .doas_megacomplex import DOASMegacomplex
from .doas_matrix import calculate_doas_matrix
from .doas_spectral_matrix import calculate_doas_spectral_matrix
from .oscillation import Oscillation


@model(
    'doas',
    attributes={
        'oscillation': Oscillation,
    },
    dataset_type=KineticSpectrumDatasetDescriptor,
    megacomplex_type=DOASMegacomplex,
    matrix=calculate_doas_matrix,
    matrix_dimension='time',
    global_matrix=calculate_doas_spectral_matrix,
    global_dimension='spectral',
    finalize_data_function=finalize_doas_data,
    constrain_matrix_function=apply_kinetic_model_constraints,
    additional_penalty_function=apply_spectral_penalties,
    grouped=grouped,
    index_dependend=index_dependend,
)
class DOASModel(KineticSpectrumModel):
    """Extends the kinetic model with damped oscillations."""



"""This package contains compartment constraint items."""

import typing
import numpy as np

from glotaran.model import model_attribute, model_attribute_typed


@model_attribute(
    properties={
        'compartment': str,
        'interval': typing.List[typing.Tuple[float, float]],
    }, has_type=True, no_label=True)
class OnlyConstraint:
    """A only constraint sets the calculated matrix row of a compartment to 0
    outside the given intervals. """

    def applies(self, index: any) -> bool:
        """
        Returns true if the indexx is in one of the intervals.

        Parameters
        ----------
        index : any

        Returns
        -------
        applies : bool

        """
        def applies(interval):
            return interval[0] <= index <= interval[1]
        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return not any([applies(i) for i in self.interval])


@model_attribute(
    properties={
        'compartment': str,
        'interval': typing.List[typing.Tuple[float, float]],
    }, has_type=True, no_label=True)
class ZeroConstraint:
    """A zero constraint sets the calculated matrix row of a compartment to 0
    in the given intervals. """

    def applies(self, index: any) -> bool:
        """
        Returns true if the indexx is in one of the intervals.

        Parameters
        ----------
        index : any

        Returns
        -------
        applies : bool

        """
        def applies(interval):
            return interval[0] <= index <= interval[1]
        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return any([applies(i) for i in self.interval])


@model_attribute_typed(types={
    'only': OnlyConstraint,
    'zero': ZeroConstraint,
}, no_label=True)
class SpectralConstraint:
    """A compartment constraint is applied on one compartment on one or many
    intervals on the estimated axies type.

    There are three types: zeroe, equal and eqal area. See the documention of
    the respective classes for details.
    """
    pass


def apply_spectral_constraints(
        model: typing.Type['KineticModel'],
        clp_labels: typing.List[str],
        matrix: np.ndarray,
        index: float) -> typing.Tuple[typing.List[str], np.ndarray]:
    for constraint in model.spectral_constraints:
        if isinstance(constraint, (OnlyConstraint, ZeroConstraint)) and constraint.applies(index):
            idx = [not label == constraint.compartment for label in clp_labels]
            clp_labels = [label for label in clp_labels if not label == constraint.compartment]
            matrix = matrix[:, idx]
    return (clp_labels, matrix)



import pytest
from glotaran.parse.parser import load_yml_file
from glotaran.builtin.models.doas import (DOASModel, DOASMegacomplex, Oscillation)

# unused import
# from glotaran.model import FixedConstraint, BoundConstraint
# from os import getcwd
from os.path import join, dirname, abspath

THIS_DIR = dirname(abspath(__file__))


@pytest.fixture
def model():
    spec_path = join(THIS_DIR, 'test_model_spec_doas.yml')
    return load_yml_file(spec_path)


def test_correct_model(model):
    assert isinstance(model, DOASModel)


def test_oscillation(model):
    assert len(model.oscillation) == 2

    i = 1
    for _ in model.oscillation:
        label = "osc{}".format(i)
        assert label in model.oscillation
        oscillation = model.oscillation[label]
        assert isinstance(oscillation, Oscillation)
        assert oscillation.label == label
        assert oscillation.frequency == i
        assert oscillation.rate == 2+i

        i = i + 1


def test_megacomplexes(model):
    assert len(model.megacomplex) == 4

    i = 1
    for _ in model.megacomplex:
        label = "cmplx{}".format(i)
        assert label in model.megacomplex
        megacomplex = model.megacomplex[label]
        assert isinstance(megacomplex, DOASMegacomplex)
        assert megacomplex.label == label
        assert megacomplex.k_matrix == ["km{}".format(i)]
        if i == 2:
            assert megacomplex.oscillation == ["osc1"]
        if i == 4:
            assert megacomplex.oscillation == ["osc2"]

        i = i + 1



import click

from . import util


@util.signature_analysis
def validate_cmd(parameter: str, model: str, scheme: str):
    """Validates a model file and optionally a parameter file."""

    if scheme is not None:
        scheme = util.load_scheme_file(scheme, verbose=True)
        click.echo(scheme.validate())
        return

    if model is not None:

        model = util.load_model_file(model, verbose=True)

        if parameter is not None:
            parameter = util.load_parameter_file(parameter, verbose=True)

        click.echo(model.validate(parameter=parameter))
    else:
        click.echo("Neither analysis scheme nor model file specified. "
                   "Type 'glotaran validate --help' for more info.")



import pytest
import numpy as np

from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel


class SimpleIrfDispersion:
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['j.1']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
            }}
        },
        'irf': {
            'irf1': {
                'type': 'spectral-gaussian',
                'center': 'irf.center',
                'width': 'irf.width',
                'dispersion_center': 'irf.dispcenter',
                'center_dispersion': ['irf.centerdisp'],
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
            },
        },
    })
    sim_model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['j.1']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
            }}
        },
        'irf': {
            'irf1': {
                'type': 'spectral-gaussian',
                'center': 'irf.center',
                'width': 'irf.width',
                'dispersion_center': 'irf.dispcenter',
                'center_dispersion': ['irf.centerdisp'],
            },
        },
        'shape': {
            'sh1': {
                'type': "one",
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
                'shape': {'s1': 'sh1'}
            },
        },
    })

    initial = ParameterGroup.from_dict({
        'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
        ],
        'kinetic': [
            ["1", 0.5],
            {'non-negative': False}
        ],
        'irf': [['center', 0.3],
                ['width', 0.1],
                ['dispcenter', 400, {'vary': False}],
                ['centerdisp', 0.5]],
    })
    wanted = ParameterGroup.from_dict({
        'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
        ],
        'kinetic': [
            ["1", 0.5],
        ],

        'irf': [['center', 0.3], ['width', 0.1], ['dispcenter', 400], ['centerdisp', 0.5]],
    })

    time_p1 = np.linspace(-1, 2, 50, endpoint=False)
    time_p2 = np.linspace(2, 5, 30, endpoint=False)
    time_p3 = np.geomspace(5, 10, num=20)
    time = np.concatenate([time_p1, time_p2, time_p3])
    spectral = np.arange(300, 500, 100)
    axis = {"time": time, "spectral": spectral}


class MultiIrfDispersion:
    model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['j.1']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
            }}
        },
        'irf': {
            'irf1': {
                'type': 'spectral-multi-gaussian',
                'center': ['irf.center'],
                'width': ['irf.width'],
                'dispersion_center': 'irf.dispcenter',
                'center_dispersion': ['irf.centerdisp1', 'irf.centerdisp2'],
                'width_dispersion': ['irf.widthdisp'],
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
            },
        },
    })
    sim_model = KineticSpectrumModel.from_dict({
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['j.1']
            },
        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {
                ("s1", "s1"): 'kinetic.1',
            }}
        },
        'irf': {
            'irf1': {
                'type': 'spectral-multi-gaussian',
                'center': ['irf.center'],
                'width': ['irf.width'],
                'dispersion_center': 'irf.dispcenter',
                'center_dispersion': ['irf.centerdisp1', 'irf.centerdisp2'],
                'width_dispersion': ['irf.widthdisp'],
            },
        },
        'shape': {
            'sh1': {
                'type': "one",
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'irf': 'irf1',
                'megacomplex': ['mc1'],
                'shape': {'s1': 'sh1'}
            },
        },
    })

    initial = ParameterGroup.from_dict({
        'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
        ],
        'kinetic': [
            ["1", 0.5],
            {'non-negative': False}
        ],
        'irf': [['center', 0.3],
                ['width', 0.1],
                ['dispcenter', 400, {'vary': False}],
                ['centerdisp1', 0.01],
                ['centerdisp2', 0.001],
                ['widthdisp', 0.025]],
    })
    wanted = ParameterGroup.from_dict({
        'j': [
            ['1', 1, {'vary': False, 'non-negative': False}],
        ],
        'kinetic': [
            ["1", 0.5],
        ],
        'irf': [['center', 0.3],
                ['width', 0.1],
                ['dispcenter', 400, {'vary': False}],
                ['centerdisp1', 0.01],
                ['centerdisp2', 0.001],
                ['widthdisp', 0.025]],
    })

    time = np.arange(-1, 5, 0.2)
    spectral = np.arange(300, 500, 100)
    axis = {"time": time, "spectral": spectral}


@pytest.mark.parametrize("suite", [
    SimpleIrfDispersion,
    MultiIrfDispersion,
])
def test_spectral_irf(suite):

    model = suite.model
    print(model.validate())
    assert model.valid()

    sim_model = suite.sim_model
    print(sim_model.validate())
    assert sim_model.valid()

    wanted = suite.wanted
    print(sim_model.validate(wanted))
    print(wanted)
    assert sim_model.valid(wanted)

    initial = suite.initial
    print(model.validate(initial))
    assert model.valid(initial)

    print(model.markdown(wanted))

    dataset = sim_model.simulate('dataset1', wanted, suite.axis)

    assert dataset.data.shape == \
        (suite.axis['time'].size, suite.axis['spectral'].size)

    data = {'dataset1': dataset}

    result = model.optimize(initial, data, nnls=True, max_nfev=20)
    print(result.optimized_parameter)

    for label, param in result.optimized_parameter.all():
        assert np.allclose(param.value, wanted.get(label).value,
                           rtol=1e-1)

    resultdata = result.data["dataset1"]

    print(resultdata)

    assert np.array_equal(dataset['time'], resultdata['time'])
    assert np.array_equal(dataset['spectral'], resultdata['spectral'])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, atol=1e-15)

    print(resultdata.fitted_data.isel(spectral=0).argmax())
    print(resultdata.fitted_data.isel(spectral=-1).argmax())
    assert resultdata.fitted_data.isel(spectral=0).argmax() != \
        resultdata.fitted_data.isel(spectral=-1).argmax()

    assert 'species_associated_spectra' in resultdata
    assert 'decay_associated_spectra' in resultdata



"""This package contains the kinetic megacomplex item."""

from typing import List

from glotaran.model import model_attribute
from glotaran.parameter import Parameter


@model_attribute(properties={
    'k_matrix': {'type': List[str], 'default': []},
    'scale': {'type': Parameter, 'allow_none': True},
})
class KineticImageMegacomplex:
    """A Megacomplex with one or more K-Matrices."""

    def full_k_matrix(self, model=None):
        full_k_matrix = None
        for k_matrix in self.k_matrix:
            if model:
                k_matrix = model.k_matrix[k_matrix]
            if full_k_matrix is None:
                full_k_matrix = k_matrix
            # If multiple k matrices are present, we combine them
            else:
                full_k_matrix = full_k_matrix.combine(k_matrix)
        return full_k_matrix

    @property
    def involved_compartments(self):
        return self.full_k_matrix().involved_compartments() if self.full_k_matrix() else []



"""Functions for simulating a global analysis model."""

import typing
import numpy as np
import xarray as xr

from glotaran.parameter import ParameterGroup


def simulate(model: typing.Type['glotaran.model.Model'],
             dataset: str,
             parameter: ParameterGroup,
             axes: typing.Dict[str, np.ndarray] = None,
             clp: typing.Union[np.ndarray, xr.DataArray] = None,
             noise=False,
             noise_std_dev=1.0,
             noise_seed=None,
             ):
    """Simulates a model.

    Parameters
    ----------
    model :
        The model to simulate.
    parameter :
        The parameters for the simulation.
    dataset :
        Label of the dataset to simulate
    axes :
        A dictory with axes for simulation.
    clp :
        Conditionaly linear parameter. Will be used instead of `model.global_matrix` if given.
    noise :
        Add noise to the simulation.
    noise_std_dev :
        The standard devition for noise simulation.
    noise_seed :
        The seed for the noise simulation.
    """

    if model.global_matrix is None and clp is None:
        raise Exception("Cannot simulate models without implementation for global matrix"
                        " and no clp given.")

    filled_dataset = model.dataset[dataset].fill(model, parameter)

    matrix_dimension = axes[model.matrix_dimension]
    global_dimension = axes[model.global_dimension]

    dim1 = matrix_dimension.size
    dim2 = global_dimension.size
    result = xr.DataArray(np.empty((dim1, dim2), dtype=np.float64),
                          coords=[(model.matrix_dimension, matrix_dimension),
                                  (model.global_dimension, global_dimension)])
    result = result.to_dataset(name='data')

    matrix = [model.matrix(dataset_descriptor=filled_dataset, axis=matrix_dimension, index=index)
              for index in global_dimension]
    if callable(model.constrain_matrix_function):
        matrix = [model.constrain_matrix_function(parameter, clp, mat, global_dimension[i])
                  for i, (clp, mat) in enumerate(matrix)]
    matrix = [xr.DataArray(mat, coords=[(model.matrix_dimension, matrix_dimension),
                                        ('clp_label', clp_label)])
              for clp_label, mat in matrix]

    if clp is not None:
        if clp.shape[0] != global_dimension.size:
            raise ValueError(f"Size of dimension 0 of clp ({clp.shape[0]}) != size of axis"
                             f" '{model.global_dimension}' ({global_dimension.size})")
        if isinstance(clp, xr.DataArray):
            if model.global_dimension not in clp.coords:
                raise ValueError(f"Missing coordinate '{model.global_dimension}' in clp.")
            if 'clp_label' not in clp.coords:
                raise ValueError(f"Missing coordinate 'clp_label' in clp.")
        else:
            if 'clp_label' not in axes:
                raise ValueError("Missing axis 'clp_label'")
            clp = xr.DataArray(clp, coords=[(model.global_dimension, global_dimension),
                                            ('clp_label', axes['clp_label'])])
    else:
        clp_labels, clp = model.global_matrix(filled_dataset, global_dimension)
        clp = xr.DataArray(clp, coords=[(model.global_dimension, global_dimension),
                                        ('clp_label', clp_labels)])
    for i in range(dim2):
        result.data[:, i] = np.dot(matrix[i], clp[i].sel(clp_label=matrix[i].coords['clp_label']))

    if noise:
        if noise_seed is not None:
            np.random.seed(noise_seed)
        result['data'] = ((model.matrix_dimension, model.global_dimension),
                          np.random.normal(result.data, noise_std_dev))

    return result




import numpy as np

from glotaran.analysis.simulation import simulate
from glotaran.parameter import ParameterGroup

from .mock import MockModel


def test_simulate_dataset():
    model = MockModel.from_dict({
        "dataset": {
            "dataset1": {
                "megacomplex": [],
            },
        }
    })
    print(model.validate())
    assert model.valid()

    parameter = ParameterGroup.from_list([1, 1])
    print(model.validate(parameter))
    assert model.valid(parameter)

    est_axis = np.asarray([1, 1, 1, 1])
    cal_axis = np.asarray([2, 2, 2])

    data = simulate(model, 'dataset1', parameter, {'e': est_axis, 'c': cal_axis})
    assert np.array_equal(data["c"], cal_axis)
    assert np.array_equal(data["e"], est_axis)
    assert data.data.shape == (3, 4)
    assert np.array_equal(data.data, np.asarray([
        [2, 4, 6],
        [4, 10, 16],
        [6, 16, 26],
        [8, 22, 36],
    ]).T)



import typing
import numpy as np
import numba as nb

from glotaran.model import model_attribute
from glotaran.builtin.models.kinetic_image.irf import Irf, IrfMultiGaussian
from glotaran.parameter import Parameter


@model_attribute(properties={
    'dispersion_center': {'type': Parameter, 'allow_none': True},
    'center_dispersion': {'type': typing.List[Parameter], 'default': []},
    'width_dispersion': {'type': typing.List[Parameter], 'default': []},
    'model_dispersion_with_wavenumber': {'type': bool, 'default': False},
}, has_type=True)
class IrfSpectralMultiGaussian(IrfMultiGaussian):
    """
    Represents a gaussian IRF.

    One width and one center is a single gauss.

    One center and multiple widths is a multiple gaussian.

    Multiple center and multiple widths is Double-, Triple- , etc. Gaussian.

    Parameters
    ----------

    label:
        label of the irf
    center:
        one or more center of the irf as parameter indices
    width:
        one or more widths of the gaussian as parameter index
    center_dispersion:
        polynomial coefficients for the dispersion of the
        center as list of parameter indices. None for no dispersion.
    width_dispersion:
        polynomial coefficients for the dispersion of the
        width as parameter indices. None for no dispersion.

    """
    def parameter(self, index):
        centers, widths, scale, backsweep, backsweep_period = \
                super(IrfSpectralMultiGaussian, self).parameter(index)

        if self.dispersion_center:
            dist = (1e3 / index - 1e3 / self.dispersion_center) \
                if self.model_dispersion_with_wavenumber else (index - self.dispersion_center)/100

        if len(self.center_dispersion) != 0:
            if self.dispersion_center is None:
                raise Exception(self, f'No dispersion center defined for irf "{self.label}"')
            for i, disp in enumerate(self.center_dispersion):
                centers += disp * np.power(dist, i+1)

        if len(self.width_dispersion) != 0:
            if self.dispersion_center is None:
                raise Exception(self, f'No dispersion center defined for irf "{self.label}"')
            for i, disp in enumerate(self.width_dispersion):
                widths = widths + disp * np.power(dist, i+1)

        return centers, widths, scale, backsweep, backsweep_period

    def calculate_dispersion(self, axis):
        dispersion = []
        for index in axis:
            center, _, _, _, _ = self.parameter(index)
            dispersion.append(center)
        return np.asarray(dispersion).T


@model_attribute(properties={
    'center': Parameter,
    'width': Parameter,
}, has_type=True)
class IrfSpectralGaussian(IrfSpectralMultiGaussian):
    pass


@model_attribute(properties={
    'coherent_artifact_order': {'type': int},
    'coherent_artifact_width': {'type': Parameter, 'allow_none': True},
}, has_type=True)
class IrfGaussianCoherentArtifact(IrfSpectralGaussian):

    def clp_labels(self):
        return [f'coherent_artifact_{i}'
                for i in range(1, self.coherent_artifact_order + 1)]

    def calculate_coherent_artifact(self, axis):
        if not 1 <= self.coherent_artifact_order <= 3:
            raise Exception(self, "Coherent artifact order must be between in [1,3]")

        center, width, _, _, _ = self.parameter(None)

        center = center[0]
        width = self.coherent_artifact_width.value \
            if self.coherent_artifact_width is not None else width[0]

        clp_label = self.clp_labels()

        matrix = self._calculate_coherent_artifact_matrix(
            center, width, axis, self.coherent_artifact_order
        )

        return clp_label, matrix

    @staticmethod
    @nb.jit(nopython=True, parallel=True)
    def _calculate_coherent_artifact_matrix(center, width, axis, order):
        matrix = np.zeros((axis.size, order), dtype=np.float64)

        matrix[:, 0] = np.exp(-1 * (axis - center)**2 / (2 * width**2))
        if order > 1:
            matrix[:, 1] = matrix[:, 0] * (center - axis) / width**2

        if order > 2:
            matrix[:, 2] = \
                matrix[:, 0] * (center**2 - width**2 - 2 * center * axis + axis**2) / width**4
        return matrix


Irf.add_type('spectral-multi-gaussian', IrfSpectralMultiGaussian)
Irf.add_type('spectral-gaussian', IrfSpectralGaussian)
Irf.add_type('gaussian-coherent-artifact', IrfGaussianCoherentArtifact)



"""This package contains the spectral shape item."""

import numpy as np
from glotaran.model import model_attribute, model_attribute_typed
from glotaran.parameter import Parameter


@model_attribute(properties={
    'amplitude': Parameter,
    'location': Parameter,
    'width': Parameter,
}, has_type=True)
class SpectralShapeGaussian:
    """A gaussian spectral shape"""

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Parameters
        ----------
        axis: np.ndarray
            The axies to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        matrix = self.amplitude * np.exp(
            -np.log(2) * np.square(2 * (axis - self.location)/self.width))
        return matrix


@model_attribute(properties={
}, has_type=True)
class SpectralShapeOne:
    """A gaussian spectral shape"""

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Parameters
        ----------
        axis: np.ndarray
            The axies to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return np.ones((axis.shape[0]))


@model_attribute(properties={
}, has_type=True)
class SpectralShapeZero:
    """A gaussian spectral shape"""

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Only works after calling calling 'fill'.

        Parameters
        ----------
        axis: np.ndarray
            The axies to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return np.zeros((axis.shape[0]))


@model_attribute_typed(types={
    'gaussian': SpectralShapeGaussian,
    'one': SpectralShapeOne,
    'zero': SpectralShapeZero,
})
class SpectralShape:
    """Base class for spectral shapes"""



import typing
import numpy as np
import xarray as xr
from scipy import fftpack

import glotaran
from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_result import \
    finalize_kinetic_spectrum_result


def finalize_doas_data(
    model: 'glotaran.models.doas.DoasModel',
    global_indices: typing.List[typing.List[object]],
    reduced_clp_labels: typing.Union[typing.Dict[str, typing.List[str]], np.ndarray],
    reduced_clps: typing.Union[typing.Dict[str, np.ndarray], np.ndarray],
    parameter: ParameterGroup, data: typing.Dict[str, xr.Dataset],
):

    finalize_kinetic_spectrum_result(
        model, global_indices, reduced_clp_labels, reduced_clps, parameter, data)

    for label in model.dataset:
        dataset = data[label]

        dataset_descriptor = model.dataset[label].fill(model, parameter)

        # get_doas

        oscillations = []

        for cmplx in dataset_descriptor.megacomplex:
            for osc in cmplx.oscillation:
                if osc.label not in oscillations:
                    oscillations.append(osc.label)

        dim1 = dataset.coords[model.global_dimension].size
        dim2 = len(oscillations)
        doas = np.zeros((dim1, dim2), dtype=np.float64)
        phase = np.zeros((dim1, dim2), dtype=np.float64)
        for i, osc in enumerate(oscillations):
            sin = dataset.clp.sel(clp_label=f'{osc}_sin')
            cos = dataset.clp.sel(clp_label=f'{osc}_cos')
            doas[:, i] = np.sqrt(sin*sin+cos*cos)
            phase[:, i] = np.unwrap(np.arctan2(sin, cos))

        dataset.coords['oscillation'] = oscillations

        dataset['dampened_oscillation_associated_spectra'] = (
            (model.global_dimension, 'oscillation'), doas)

        dataset['dampened_oscillation_phase'] = (
            (model.global_dimension, 'oscillation'), phase)

        dataset['dampened_oscillation_sin'] = \
            dataset.matrix.sel(clp_label=[f'{osc}_sin' for osc in oscillations])\
            .rename(clp_label='oscillation')

        dataset['dampened_oscillation_cos'] = \
            dataset.matrix.sel(clp_label=[f'{osc}_cos' for osc in oscillations])\
            .rename(clp_label='oscillation')

    time_diff = np.diff(dataset.time, n=1, axis=0)

    power = dataset.residual_left_singular_vectors.isel(left_singular_value_index=0).values[:-1]
    power = power[time_diff < time_diff.mean()]

    power = fftpack.fft(power, n=1024, axis=0)

    power = np.abs(power)/power.size

    dataset['residual_power_spectrum'] = (('frequency'), power)



from glotaran.model import model, Model

from .initial_concentration import InitialConcentration
from .irf import Irf
from .k_matrix import KMatrix
from .kinetic_image_dataset_descriptor import KineticImageDatasetDescriptor
from .kinetic_image_matrix import kinetic_image_matrix
from .kinetic_image_megacomplex import KineticImageMegacomplex
from .kinetic_image_result import finalize_kinetic_image_result


@model(
    'kinetic-image',
    attributes={
        'initial_concentration': InitialConcentration,
        'k_matrix': KMatrix,
        'irf': Irf,
    },
    dataset_type=KineticImageDatasetDescriptor,
    megacomplex_type=KineticImageMegacomplex,
    matrix=kinetic_image_matrix,
    matrix_dimension='time',
    global_dimension='pixel',
    grouped=False,
    finalize_data_function=finalize_kinetic_image_result,
)
class KineticImageModel(Model):
    pass



"""Glotaran DOAS Model Spectral Matrix"""

import numpy as np

from glotaran.builtin.models.kinetic_spectrum.spectral_matrix import spectral_matrix

from .doas_matrix import _collect_oscillations


def calculate_doas_spectral_matrix(dataset, axis):

    oscillations = _collect_oscillations(dataset)
    all_oscillations = []
    clp = []
    for osc in oscillations:
        all_oscillations.append(osc)
        all_oscillations.append(osc)
        clp.append(f'{osc.label}_sin')
        clp.append(f'{osc.label}_cos')
    matrix = np.ones((axis.size, len(all_oscillations)), dtype=np.float64)
    for i, osc in enumerate(all_oscillations):
        if osc.label not in dataset.shape:
            raise Exception(f'No shape for oscillation "{osc.label}"')
        shapes = dataset.shape[osc.label]
        if not isinstance(shapes, list):
            shapes = [shapes]
        for shape in shapes:
            matrix[:, i] *= shape.calculate(axis)

    kinetic_spectral_clp, kinetic_spectral_matrix = spectral_matrix(dataset, axis)

    if kinetic_spectral_matrix is not None:
        matrix = np.concatenate((matrix, kinetic_spectral_matrix), axis=1)
        clp += kinetic_spectral_clp

    return clp, matrix



import typing

from glotaran.model import model_attribute
from glotaran.builtin.models.kinetic_image.kinetic_image_dataset_descriptor \
    import KineticImageDatasetDescriptor


@model_attribute(properties={
    'shape': {'type': typing.Dict[str, str], 'allow_none': True},
})
class KineticSpectrumDatasetDescriptor(KineticImageDatasetDescriptor):
    pass



"""Glotaran Spectral Matrix"""

import numpy as np


def spectral_matrix(dataset, axis):
    """ Calculates the matrix.

    Parameters
    ----------
    matrix : np.array
        The preallocated matrix.

    compartment_order : list(str)
        A list of compartment labels to map compartments to indices in the
        matrix.

    parameter : glotaran.model.ParameterGroup

    """
    if dataset.initial_concentration is None:
        return None, None
    shape_compartments = [s for s in dataset.shape]
    compartments = [c for c in dataset.initial_concentration.compartments
                    if c in shape_compartments]
    matrix = np.zeros((axis.size, len(compartments)))
    for i, comp in enumerate(compartments):
        shapes = dataset.shape[comp]
        if not isinstance(shapes, list):
            shapes = [shapes]
        for shape in shapes:
            matrix[:, i] += shape.calculate(axis)
    return compartments, matrix



"""Helper functions."""

import typing


def wrap_func_as_method(cls: typing.Any,
                        name: str = None,
                        annotations: str = None,
                        doc: str = None) -> typing.Callable:
    """A decorator to wrap a function as class method.

    Notes
    -----

    Only for internal use.

    Parameters
    ----------
    cls :
        The class in which the function will be wrapped.
    name :
        The name of method. If `None`, the orginal function's name is used.
    annotations :
        The annotations of the method. If `None`, the orginal function's annotations are used.
    doc :
        The documenation of the method. If `None`, the orginal function's documenation is used.
    """

    def decorator(func):
        if name:
            func.__name__ = name
        if annotations:
            setattr(func, '__annotations__', annotations)
        if doc:
            func.__doc__ = doc
        func.__qualname__ = cls.__qualname__ + '.' + func.__name__
        func.__module__ = cls.__module__

        return func

    return decorator



import numpy as np
import pandas as pd
import xarray as xr
import pytest

from glotaran.builtin.file_formats.sdt.sdt_file_reader import read_sdt
from . import TEMPORAL_DATA


@pytest.mark.parametrize("test_file_path, result_file_path, index", [
    (TEMPORAL_DATA["sdt"], TEMPORAL_DATA["csv"], [1]),
])
def test_read_sdt(test_file_path, result_file_path, index):

    test_dataset = read_sdt(file_path=test_file_path, index=index)
    result_df = pd.read_csv(result_file_path, skiprows=1, sep=r"\s+",
                            dtype={"Delay": np.float, "Data": np.uint16})
    result_df.Delay = result_df.Delay * 1e-9
    result_traces = pd.DataFrame([result_df.Data.values], columns=result_df.Delay)
    assert isinstance(test_dataset, xr.Dataset)
    assert np.all(test_dataset.data.T[0] == result_df.Data.values)

    assert test_dataset.data.T.shape == result_traces.values.shape
    assert np.allclose(test_dataset.time, np.array(result_traces.columns))



"""The parameter class."""

import typing
import numpy as np

from lmfit import Parameter as LmParameter

import glotaran


class Keys:
    """Keys for parameter options."""
    MIN = "min"
    MAX = "max"
    NON_NEG = "non-negative"
    EXPR = "expr"
    VARY = "vary"


class Parameter(LmParameter):
    """A parameter for optimization."""

    def __init__(self, label: str = None, full_label: str = None):
        """

        Parameters
        ----------
        label :
            The label of the parameter.
        full_label : str
            The label of the parameter with its path in a paramter group prepended.
        """

        super(Parameter, self).__init__(
            name=label, user_data={'non_neg': False, 'full_label': full_label})

        self.label = label
        self.full_label = full_label

    @classmethod
    def from_parameter(cls, label: str, parameter: LmParameter) -> 'Parameter':
        """Creates a :class:`Parameter` from a `lmfit.Parameter`

        Parameters
        ----------
        label : str
            The label of the parameter.
        parameter : lmfit.Parameter
            The `lmfit.Parameter`.
        """
        p = cls(label=label)
        p.vary = parameter.vary
        p.non_neg = parameter.user_data['non_neg']
        p.full_label = parameter.user_data['full_label']
        p.value = np.exp(parameter.value) if p.non_neg else parameter.value
        p.min = \
            np.exp(parameter.min) if p.non_neg and np.isfinite(parameter.min) else parameter.min
        p.max = \
            np.exp(parameter.max) if p.non_neg and np.isfinite(parameter.max) else parameter.max
        p.stderr = parameter.stderr
        return p

    @classmethod
    def from_list_or_value(cls,
                           value: typing.Union[int, float, typing.List],
                           default_options: typing.Dict = None,
                           label: str = None) -> 'Parameter':
        """Creates a parameter from a list or numeric value.

        Parameters
        ----------
        value :
            The list or numeric value.
        default_options :
            A dictonary of default options.
        label :
            The label of the parameter.
        """

        if not isinstance(value, list):
            value = [value]

        param = cls(label=label)

        if isinstance(value, (int, float)):
            param.value = value
            return param

        def retrieve(filt, default):
            tmp = list(filter(filt, value))
            if len(tmp) != 0:
                value.remove(tmp[0])
                return tmp[0]
            else:
                return default
        options = retrieve(lambda x: isinstance(x, dict), None)

        param.label = value[0] if len(value) != 1 else label
        param.value = float(value[0] if len(value) == 1 else value[1])

        if default_options:
            param._set_options_from_dict(default_options)

        if options:
            param._set_options_from_dict(options)
        return param

    def set_from_group(self, group: 'glotaran.parameter.ParameterGroup'):
        """Sets all values of the parameter to the values of the conrresoping parameter in the group.

        Notes
        -----

        For internal use.

        Parameters
        ----------
        group :
            The :class:`glotaran.parameter.ParameterGroup`.
        """

        p = group.get(self.full_label)
        self.vary = p.vary
        self.value = p.value
        self.min = p.min
        self.max = p.max
        self.expr = p.expr
        self.stderr = p.stderr
        self.non_neg = p.non_neg

    def _set_options_from_dict(self, options: typing.Dict):
        if Keys.NON_NEG in options:
            self.non_neg = options[Keys.NON_NEG]
        if Keys.MAX in options:
            self.max = float(options[Keys.MAX])
        if Keys.MIN in options:
            self.min = float(options[Keys.MIN])
        if Keys.EXPR in options:
            self.expr = options[Keys.EXPR]
        if Keys.VARY in options:
            self.vary = options[Keys.VARY]

    @property
    def label(self) -> str:
        """Label of the parameter"""
        return self.user_data['label']

    @label.setter
    def label(self, label: str):
        self.user_data['label'] = label

    @property
    def full_label(self) -> str:
        """The label of the parameter with its path in a parameter group prepended."""
        return self.user_data['full_label']

    @full_label.setter
    def full_label(self, full_label: str):
        self.user_data['full_label'] = full_label

    @property
    def non_neg(self) -> bool:
        """Indicates if the parameter is non-negativ.

        If true, the parameter will be transformed with :math:`p' = \log{p}` and
        :math:`p = \exp{p'}`.
        """  # noqa w605
        return self.user_data['non_neg']

    @non_neg.setter
    def non_neg(self, non_neg: bool):
        self.user_data['non_neg'] = non_neg

    @LmParameter.value.setter
    def value(self, val):
        if not isinstance(val, (int, float)):
            try:
                val = float(val)
            except Exception:
                raise Exception("Parameter Error: value must be numeric:"
                                "{} Type: {}".format(val, type(val)))

        if isinstance(val, int):
            val = float(val)

        LmParameter.value.fset(self, val)

    def __str__(self):
        """ """
        return f"__{self.label}__: _Value_: {self.value}, _StdErr_: {self.stderr}, _Min_:" + \
               f" {self.min}, _Max_: {self.max}, _Vary_: {self.vary}," + \
               f" _Non-Negative_: {self.non_neg}"



import typing
import numpy as np
import xarray as xr

import glotaran
from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.builtin.models.kinetic_image.kinetic_image_result import (
    retrieve_species_assocatiated_data,
    retrieve_decay_assocatiated_data,
    retrieve_irf,
)

from .spectral_irf import IrfSpectralMultiGaussian, IrfGaussianCoherentArtifact
from .spectral_constraints import OnlyConstraint, ZeroConstraint


def finalize_kinetic_spectrum_result(
    model: 'glotaran.models.spectral_temporal.KineticModel',
    global_indices: typing.List[typing.List[object]],
    reduced_clp_labels: typing.Union[typing.Dict[str, typing.List[str]], np.ndarray],
    reduced_clps: typing.Union[typing.Dict[str, np.ndarray], np.ndarray],
    parameter: ParameterGroup, data: typing.Dict[str, xr.Dataset],
):

    for label in model.dataset:
        dataset = data[label]
        dataset_descriptor = model.dataset[label].fill(model, parameter)

        if not dataset_descriptor.get_k_matrices():
            continue

        retrieve_species_assocatiated_data(model, dataset, dataset_descriptor, "spectra")

        if dataset_descriptor.baseline:
            dataset['baseline'] = dataset.clp.sel(clp_label=f"{dataset_descriptor.label}_baseline")

        for constraint in model.spectral_constraints:
            if isinstance(constraint, (OnlyConstraint, ZeroConstraint)):
                idx = [index for index in dataset.spectral if constraint.applies(index)]

        for relation in model.spectral_relations:
            if relation.compartment in dataset.coords['species']:
                relation = relation.fill(model, parameter)

                # indexes on the global axis
                idx = [index for index in dataset.spectral if relation.applies(index)]
                dataset.species_associated_spectra\
                    .loc[{'species': relation.target, model.global_dimension: idx}] = \
                    dataset.species_associated_spectra.sel(
                        {'species': relation.compartment, model.global_dimension: idx}) * \
                    relation.parameter

        retrieve_decay_assocatiated_data(model, dataset, dataset_descriptor, "spectra")

        irf = dataset_descriptor.irf
        if isinstance(irf, IrfMultiGaussian):
            if isinstance(irf.center, list):
                dataset['irf_center'] = irf.center[0].value
                dataset['irf_width'] = irf.width[0].value
            else:
                dataset['irf_center'] = irf.center.value
                dataset['irf_width'] = irf.width.value
        if isinstance(irf, IrfSpectralMultiGaussian):
            index = irf.dispersion_center if irf.dispersion_center \
                 else dataset.coords[model.global_dimension].min().values
            dataset['irf'] = (('time'), irf.calculate(index, dataset.coords['time']))

            if irf.dispersion_center:
                for i, dispersion in enumerate(
                        irf.calculate_dispersion(dataset.coords['spectral'].values)):
                    dataset[f'center_dispersion_{i+1}'] = \
                            ((model.global_dimension, dispersion))
        if isinstance(irf, IrfGaussianCoherentArtifact):
            dataset.coords['coherent_artifact_order'] = \
                    list(range(1, irf.coherent_artifact_order+1))
            dataset['coherent_artifact_concentration'] = (
                (
                    model.matrix_dimension,
                    'coherent_artifact_order'),
                dataset.matrix.sel(clp_label=irf.clp_labels()).values
            )
            dataset['coherent_artifact_associated_spectra'] = (
                (model.global_dimension, 'coherent_artifact_order'),
                dataset.clp.sel(clp_label=irf.clp_labels()).values
            )

        else:
            retrieve_irf(model, dataset, dataset_descriptor, "images")



import collections
import dask
import numpy as np


LabelAndMatrix = collections.namedtuple('LabelAndMatrix', 'clp_label matrix')
LabelAndMatrixAndData = collections.namedtuple('LabelAndMatrixAndData', 'label_matrix data')


def calculate_index_independend_ungrouped_matrices(scheme, parameter):

    # for this we don't use dask

    clp_labels = {}
    matrices = {}
    constraint_labels_and_matrices = {}
    model = scheme.model

    descriptors = {label: descriptor.fill(model, parameter)
                   for label, descriptor in scheme.model.dataset.items()}

    for label, descriptor in descriptors.items():
        axis = scheme.data[label].coords[model.matrix_dimension].values
        clp_label, matrix = _calculate_matrix(
            model.matrix,
            descriptor,
            axis,
            {},
        )
        clp_labels[label] = clp_label
        matrices[label] = matrix

        if callable(model.has_matrix_constraints_function):
            if model.has_matrix_constraints_function():
                clp_label, matrix = \
                    model.constrain_matrix_function(parameter, clp_label, matrix, None)

        constraint_labels_and_matrices[label] = LabelAndMatrix(clp_label, matrix)
    return clp_labels, matrices, constraint_labels_and_matrices


def calculate_index_independend_grouped_matrices(scheme, groups, parameter):

    # We just need to create groups from the ungrouped matrices
    clp_labels, matrices, constraint_labels_and_matrices = \
        calculate_index_independend_ungrouped_matrices(scheme, parameter)
    for label, group in groups.items():
        if label not in matrices:
            constraint_labels_and_matrices[label] = \
                _combine_matrices([constraint_labels_and_matrices[l] for l in group])

    return clp_labels, matrices, constraint_labels_and_matrices


def create_index_dependend_ungrouped_matrix_jobs(scheme, bag, parameter):

    model = scheme.model
    clp_labels = {}
    matrices = {}
    constraint_labels_and_matrices = {}

    descriptors = {label: descriptor.fill(model, parameter)
                   for label, descriptor in scheme.model.dataset.items()}
    for label, problem in bag.items():
        descriptor = descriptors[label]
        clp_labels[label] = []
        matrices[label] = []
        constraint_labels_and_matrices[label] = []
        for index in problem.global_axis:
            clp, matrix = dask.delayed(_calculate_matrix, nout=2)(
                model.matrix,
                descriptor,
                problem.matrix_axis,
                {},
                index=index,
            )
            clp_labels[label].append(clp)
            matrices[label].append(matrix)

            if callable(model.has_matrix_constraints_function):
                if model.has_matrix_constraints_function():
                    clp, matrix = dask.delayed(model.constrain_matrix_function, nout=2)(
                        parameter, clp, matrix, index
                    )
            constraint_labels_and_matrices[label].append((clp, matrix))

    return clp_labels, matrices, constraint_labels_and_matrices


def create_index_dependend_grouped_matrix_jobs(scheme, bag, parameter):

    model = scheme.model

    descriptors = {label: descriptor.fill(model, parameter)
                   for label, descriptor in scheme.model.dataset.items()}

    def calculate_group(group):
        results = [_calculate_matrix(
            model.matrix,
            descriptors[problem.dataset],
            problem.axis,
            {},
            index=problem.index
        ) for problem in group.descriptor]
        return results, group.descriptor[0].index

    def get_clp(result):
        return [d[0] for d in result[0]]

    def get_matrices(result):
        return [d[1] for d in result[0]]

    def constrain_and_combine_matrices(result):
        matrices, index = result
        clp, matrix = _combine_matrices(matrices)
        if callable(model.has_matrix_constraints_function):
            if model.has_matrix_constraints_function():
                clp, matrix = model.constrain_matrix_function(parameter, clp, matrix, index)
        return LabelAndMatrix(clp, matrix)

    matrix_jobs = bag.map(calculate_group)
    constraint_matrix_jobs = matrix_jobs.map(constrain_and_combine_matrices)
    clp = matrix_jobs.map(get_clp)
    matrices = matrix_jobs.map(get_matrices)
    return clp, matrices, constraint_matrix_jobs


def _calculate_matrix(matrix_function, dataset_descriptor, axis, extra, index=None):
    args = {
        'dataset_descriptor': dataset_descriptor,
        'axis': axis,
    }
    for k, v in extra:
        args[k] = v
    if index is not None:
        args['index'] = index
    clp_label, matrix = matrix_function(**args)
    if dataset_descriptor.scale is not None:
        matrix *= dataset_descriptor.scale
    return LabelAndMatrix(clp_label, matrix)


def _combine_matrices(labels_and_matrices):
    masks = []
    full_clp_labels = None
    sizes = []
    for label_and_matrix in labels_and_matrices:
        (clp_label, matrix) = label_and_matrix
        sizes.append(matrix.shape[0])
        if full_clp_labels is None:
            full_clp_labels = clp_label
            masks.append([i for i, _ in enumerate(clp_label)])
        else:
            mask = []
            for c in clp_label:
                if c not in full_clp_labels:
                    full_clp_labels.append(c)
                mask.append(full_clp_labels.index(c))
            masks.append(mask)
    dim1 = np.sum(sizes)
    dim2 = len(full_clp_labels)
    full_matrix = np.zeros((dim1, dim2), dtype=np.float64)
    start = 0
    for i, m in enumerate(labels_and_matrices):
        end = start + sizes[i]
        full_matrix[start:end, masks[i]] = m[1]
        start = end

    return (full_clp_labels, full_matrix)



import numpy as np
import xarray as xr

from glotaran.parameter import ParameterGroup
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_matrix \
    import kinetic_spectrum_matrix


def test_coherent_artifact():
    model_dict = {
        'initial_concentration': {
            'j1': {
                'compartments': ['s1'],
                'parameters': ['2']
            },

        },
        'megacomplex': {
            'mc1': {'k_matrix': ['k1']},
        },
        'k_matrix': {
            "k1": {'matrix': {("s1", "s1"): '1', }}
        },
        'irf': {
            'irf1': {
                'type': 'gaussian-coherent-artifact',
                'center': '2',
                'width': '3',
                'coherent_artifact_order': 3,
            },
        },
        'dataset': {
            'dataset1': {
                'initial_concentration': 'j1',
                'megacomplex': ['mc1'],
                'irf': 'irf1',
            },
        },
    }
    model = KineticSpectrumModel.from_dict(model_dict.copy())

    parameter = ParameterGroup.from_list([
        101e-4,
        [10, {'vary': False, 'non-negative': False}],
        [20, {'vary': False, 'non-negative': False}],
        [30, {'vary': False, 'non-negative': False}],
    ])

    time = np.asarray(np.arange(0, 50, 1.5))

    irf = model.irf['irf1'].fill(model, parameter)
    irf_same_width = irf.calculate_coherent_artifact(time)

    model_dict['irf']['irf1']['coherent_artifact_width'] = '4'
    model = KineticSpectrumModel.from_dict(model_dict)

    irf = model.irf['irf1'].fill(model, parameter)
    irf_diff_width = irf.calculate_coherent_artifact(time)

    assert not np.array_equal(irf_same_width, irf_diff_width)

    dataset = model.dataset['dataset1'].fill(model, parameter)
    compartments, matrix = kinetic_spectrum_matrix(dataset, time, 0)

    assert len(compartments) == 4
    for i in range(1, 4):
        assert compartments[i] == f'coherent_artifact_{i}'

    assert matrix.shape == (time.size, 4)

    clp = xr.DataArray([[1, 1, 1, 1]],
                       coords=[('spectral', [0]),
                               ('clp_label', ['s1',
                                              'coherent_artifact_1',
                                              'coherent_artifact_2',
                                              'coherent_artifact_3',
                                              ])])
    axis = {"time": time, "spectral": clp.spectral}
    dataset = model.simulate('dataset1', parameter, axis, clp)

    data = {'dataset1': dataset}
    result = model.optimize(parameter, data, max_nfev=20)
    print(result.optimized_parameter)

    for label, param in result.optimized_parameter.all():
        assert np.allclose(param.value, parameter.get(label).value,
                           rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset['time'], resultdata['time'])
    assert np.array_equal(dataset['spectral'], resultdata['spectral'])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)

    assert 'coherent_artifact_concentration' in resultdata
    assert resultdata['coherent_artifact_concentration'].shape == (time.size, 3)

    assert 'coherent_artifact_associated_spectra' in resultdata
    assert resultdata['coherent_artifact_associated_spectra'].shape == (1, 3)




import click

import glotaran as gta

from .commands.print import print_cmd
from .commands.validate import validate_cmd
from .commands.optimize import optimize_cmd
from .commands.pluginlist import plugin_list_cmd


class Cli(click.Group):

    def __init__(self, *args, **kwargs):
        self.help_priorities = {}
        super(Cli, self).__init__(*args, **kwargs)

    def get_help(self, ctx):
        self.list_commands = self.list_commands_for_help
        return super(Cli, self).get_help(ctx)

    def list_commands_for_help(self, ctx):
        """reorder the list of commands when listing the help"""
        commands = super(Cli, self).list_commands(ctx)
        return (c[1] for c in sorted(
            (self.help_priorities.get(command, 1), command)
            for command in commands))

    def command(self, *args, **kwargs):
        """Behaves the same as `click.Group.command()` except capture
        a priority for listing command names in help.
        """
        help_priority = kwargs.pop('help_priority', 1)
        help_priorities = self.help_priorities

        def decorator(f):
            cmd = super(Cli, self).command(*args, **kwargs)(f)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator


@click.group(cls=Cli)
@click.version_option(version=gta.__version__)
def glotaran():
    pass


glotaran.add_command(glotaran.command(
    name='pluginlist',
    short_help='Prints a list of installed plugins.', help_priority=4)(plugin_list_cmd))
glotaran.add_command(glotaran.command(
    name='print', short_help="Prints a model as markdown.", help_priority=3)(print_cmd))
glotaran.add_command(glotaran.command(
    name='validate', short_help='Validates a model file.', help_priority=2)(validate_cmd))
glotaran.add_command(glotaran.command(
    name='optimize', short_help="Optimizes a model.", help_priority=1)(optimize_cmd))


if __name__ == '__main__':
    glotaran()



""" Glotaran Spectral Relation """

import typing
import numpy as np

from glotaran.model import model_attribute
from glotaran.parameter import Parameter, ParameterGroup


@model_attribute(
    properties={
        'compartment': str,
        'target': str,
        'parameter': Parameter,
        'interval': typing.List[typing.Tuple[float, float]],
    }, no_label=True)
class SpectralRelation:
    def applies(self, index: any) -> bool:
        """
        Returns true if the index is in one of the intervals.

        Parameters
        ----------
        index : any

        Returns
        -------
        applies : bool

        """
        return any([interval[0] <= index <= interval[1] for interval in self.interval])


def create_spectral_relation_matrix(
        model: typing.Type['KineticModel'],
        parameter: ParameterGroup,
        clp_labels: typing.List[str],
        matrix: np.ndarray,
        index: float) -> typing.Tuple[typing.List[str], np.ndarray]:
    relation_matrix = np.diagflat([1.0 for _ in clp_labels])

    idx_to_delete = []
    for relation in model.spectral_relations:
        if relation.compartment in clp_labels and relation.applies(index):
            relation = relation.fill(model, parameter)
            source_idx = clp_labels.index(relation.compartment)
            target_idx = clp_labels.index(relation.target)
            relation_matrix[target_idx, source_idx] = relation.parameter
            idx_to_delete.append(target_idx)

    clp_labels = [label for i, label in enumerate(clp_labels) if i not in idx_to_delete]
    relation_matrix = np.delete(relation_matrix, idx_to_delete, axis=1)
    return (clp_labels, relation_matrix)


def apply_spectral_relations(
        model: typing.Type['KineticModel'],
        parameter: ParameterGroup,
        clp_labels: typing.List[str],
        matrix: np.ndarray,
        index: float) -> typing.Tuple[typing.List[str], np.ndarray]:

    if not model.spectral_relations:
        return (clp_labels, matrix)

    reduced_clp_labels, relation_matrix = \
        create_spectral_relation_matrix(model, parameter, clp_labels, matrix, 1)

    reduced_matrix = matrix @ relation_matrix

    return (reduced_clp_labels, reduced_matrix)


def retrieve_clps(
        model: typing.Type['KineticModel'],
        parameter: ParameterGroup,
        reduced_clp_labels: typing.List[str],
        reduced_clps: np.ndarray,
        index: float) -> typing.Tuple[typing.List[str], np.ndarray]:

    if not model.spectral_relations:
        return reduced_clp_labels, reduced_clps

    retrieved_clp_labels = []
    retrieved_clps = []

    for relation in model.spectral_relations:
        if relation.compartment in reduced_clp_labels and relation.applies(index):
            relation = relation.fill(model, parameter)
            retrieved_clp_labels.append(relation.target)
            source_idx = reduced_clp_labels.index(relation.compartment)
            retrieved_clps.append(
                reduced_clps[source_idx] * relation.parameter
            )

    retrieved_clps = \
        np.concatenate([reduced_clps, retrieved_clps]) if retrieved_clps else reduced_clps

    return reduced_clp_labels + retrieved_clp_labels, retrieved_clps



"""Functions for reading and parsing models from serialized representations."""

from typing import Dict
import yaml
import re
from .register import get_model, known_model


# shamelessly taken from
# https://stackoverflow.com/questions/39553008/how-to-read-a-python-tuple-using-pyyaml#39553138
# this is to convert the string written as a tuple into a python tuple
def _yml_tuple_constructor(loader, node):
    # this little parse is really just for what I needed, feel free to change it!
    def parse_tup_el(el):
        # try to convert into int or float else keep the string
        if el.isdigit():
            return int(el)
        try:
            return float(el)
        except ValueError:
            return el

    value = loader.construct_scalar(node)
    # remove the ( ) from the string
    tup_elements = [ele.strip() for ele in value[1:-1].split(',')]
    # remove the last element if the tuple was written as (x,b,)
    if tup_elements[-1] == '':
        tup_elements.pop(-1)
    tup = tuple(map(parse_tup_el, tup_elements))
    return tup


# !tuple is my own tag name, I think you could choose anything you want
yaml.FullLoader.add_constructor(u'!tuple', _yml_tuple_constructor)
# this is to spot the strings written as tuple in the yaml
yaml.FullLoader.add_implicit_resolver(u'!tuple', re.compile(r"\((.*?,.*?)\)"), None)


def parse_yml_file(fname: str) -> Dict:
    """parse_yml_file reads the given file and parses its content as YML.

    Parameters
    ----------
    fname : str
        fname is the of the file to parse.

    Returns
    -------
    content : Dict
        The content of the file as dictionary.
    """

    with open(fname) as f:
        spec = parse_yml(f)

    return spec


def parse_yml(data: str):
    try:
        return yaml.load(data, Loader=yaml.FullLoader)
    except Exception as e:
        raise e


def parse_spec(spec: Dict):

    if 'type' not in spec:
        raise Exception("Model type not defined")

    model_type = spec['type']
    del spec['type']

    if not known_model(model_type):
        raise Exception(f"Unknown model type '{model_type}'.")

    model = get_model(model_type)
    try:
        return model.from_dict(spec)
    except Exception as e:
        raise e


def load_yml_file(fname: str):
    spec = parse_yml_file(fname)
    return parse_spec(spec)


def load_yml(data: str):
    spec = parse_yml(data)
    return parse_spec(spec)
